import collections
import datetime
import glob
import inspect
import json
import os
import re

import keras
from absl import logging

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.utils import tensor_utils
from keras_hub.src.utils.keras_utils import print_msg
from keras_hub.src.utils.keras_utils import sharded_weights_available
from keras_hub.src.utils.tensor_utils import get_tensor_size_in_bits

try:
    import kagglehub
    from kagglehub.exceptions import KaggleApiHTTPError
except ImportError:
    kagglehub = None

try:
    import huggingface_hub
    from huggingface_hub.utils import EntryNotFoundError
    from huggingface_hub.utils import HFValidationError
except ImportError:
    huggingface_hub = None

KAGGLE_PREFIX = "kaggle://"
GS_PREFIX = "gs://"
HF_PREFIX = "hf://"
MODELSCOPE_PREFIX = "modelscope://"

KAGGLE_SCHEME = "kaggle"
GS_SCHEME = "gs"
HF_SCHEME = "hf"
MODELSCOPE_SCHEME = "modelscope"
ASSET_DIR = "assets"
TOKENIZER_ASSET_DIR = f"{ASSET_DIR}/tokenizer"

# Config file names.
CONFIG_FILE = "config.json"
TOKENIZER_CONFIG_FILE = "tokenizer.json"
AUDIO_CONVERTER_CONFIG_FILE = "audio_converter.json"
IMAGE_CONVERTER_CONFIG_FILE = "image_converter.json"
TASK_CONFIG_FILE = "task.json"
PREPROCESSOR_CONFIG_FILE = "preprocessor.json"
METADATA_FILE = "metadata.json"

# Weight file names.
MODEL_WEIGHTS_FILE = "model.weights.h5"
TASK_WEIGHTS_FILE = "task.weights.h5"
SHARDED_MODEL_WEIGHTS_CONFIG_FILE = "model.weights.json"

# HuggingFace filenames.
README_FILE = "README.md"
HF_CONFIG_FILE = "config.json"
HF_TOKENIZER_CONFIG_FILE = "tokenizer_config.json"
SAFETENSOR_CONFIG_FILE = "model.safetensors.index.json"
SAFETENSOR_FILE = "model.safetensors"

# Global state for preset registry.
BUILTIN_PRESETS = {}
BUILTIN_PRESETS_FOR_BACKBONE = collections.defaultdict(dict)


def register_presets(presets, backbone_cls):
    """Register built-in presets for a set of classes.

    Note that this is intended only for models and presets shipped in the
    library itself.
    """
    for preset in presets:
        BUILTIN_PRESETS[preset] = presets[preset]
        BUILTIN_PRESETS_FOR_BACKBONE[backbone_cls][preset] = presets[preset]


def builtin_presets(cls):
    """Find all registered built-in presets for a class."""
    presets = {}
    if cls in BUILTIN_PRESETS_FOR_BACKBONE:
        presets.update(BUILTIN_PRESETS_FOR_BACKBONE[cls])
    backbone_cls = getattr(cls, "backbone_cls", None)
    if backbone_cls:
        presets.update(builtin_presets(backbone_cls))
    for subclass in list_subclasses(cls):
        presets.update(builtin_presets(subclass))
    return presets


def list_subclasses(cls):
    """Find all registered subclasses of a class."""
    # Deduplicate the lists, since we have to register object twice for compat.
    custom_objects = set(keras.saving.get_custom_objects().values())
    subclasses = []
    for x in custom_objects:
        if inspect.isclass(x) and x != cls and issubclass(x, cls):
            subclasses.append(x)
    return subclasses


def find_subclass(preset, cls, backbone_cls):
    """Find a subclass that is compatible with backbone_cls."""
    subclasses = list_subclasses(cls)
    subclasses = filter(lambda x: x.backbone_cls == backbone_cls, subclasses)
    subclasses = list(subclasses)
    if not subclasses:
        raise ValueError(
            f"Unable to find a subclass of {cls.__name__} that is compatible "
            f"with {backbone_cls.__name__} found in preset '{preset}'."
        )
    # If we find multiple subclasses, try to filter to direct subclasses of
    # the class we are trying to instantiate.
    if len(subclasses) > 1:
        directs = list(filter(lambda x: x in cls.__bases__, subclasses))
        if len(directs) > 1:
            subclasses = directs
    # Return the subclass that was registered first (prefer built-in classes).
    return subclasses[0]


def get_file(preset, path):
    """Download a preset file in necessary and return the local path."""
    # TODO: Add tests for FileNotFound exceptions.
    if not isinstance(preset, str):
        raise ValueError(
            f"A preset identifier must be a string. Received: preset={preset}"
        )
    if preset in BUILTIN_PRESETS:
        preset = BUILTIN_PRESETS[preset]["kaggle_handle"]

    scheme = None
    if "://" in preset:
        scheme = preset.split("://")[0].lower()

    if scheme == KAGGLE_SCHEME:
        if kagglehub is None:
            raise ImportError(
                "`from_preset()` requires the `kagglehub` package. "
                "Please install with `pip install kagglehub`."
            )
        kaggle_handle = preset.removeprefix(KAGGLE_SCHEME + "://")
        num_segments = len(kaggle_handle.split("/"))
        if num_segments not in (4, 5):
            raise ValueError(
                "Unexpected Kaggle preset. Kaggle model handles should have "
                "the form kaggle://{org}/{model}/keras/{variant}[/{version}]. "
                "For example, 'kaggle://username/bert/keras/bert_base_en' or "
                "'kaggle://username/bert/keras/bert_base_en/1' (to specify a "
                f"version). Received: preset={preset}"
            )
        try:
            return kagglehub.model_download(kaggle_handle, path)
        except KaggleApiHTTPError as e:
            message = str(e)
            if message.find("403 Client Error"):
                raise FileNotFoundError(
                    f"`{path}` doesn't exist in preset directory `{preset}`."
                )
            else:
                raise ValueError(message)
        except ValueError as e:
            message = str(e)
            if message.find("is not present in the model files"):
                raise FileNotFoundError(
                    f"`{path}` doesn't exist in preset directory `{preset}`."
                )
            else:
                raise ValueError(message)
    elif scheme in tf_registered_schemes():
        return tf_copy_gfile_to_cache(preset, path)
    elif scheme == MODELSCOPE_SCHEME:
        try:
            from modelscope.hub.snapshot_download import snapshot_download
        except ImportError:
            raise ImportError(
                "To load a preset from ModelScope {preset} using from_preset,"
                "install the modelscope package with: pip install modelscope."
            )
        modelscope_handle = preset.removeprefix(MODELSCOPE_SCHEME + "://")
        try:
            return_path = snapshot_download(modelscope_handle) + "/" + path
            if os.path.exists(return_path):
                return return_path
            raise FileNotFoundError(
                f"`{return_path}` doesn't exist in preset directory `{preset}`."
            )
        except ValueError as e:
            raise ValueError(
                "ModelScope handles should follow the format "
                f"'modelscope://{{org}}/{{model}}'  "
                "(e.g., 'modelscope://username/bert_base_en')."
                f"Received: preset='{preset}.'"
            ) from e
        except EntryNotFoundError as e:
            message = str(e)
            if message.find("403 Client Error"):
                raise FileNotFoundError(
                    f"`{path}` not exist in preset directory `{preset}`."
                )
            else:
                raise ValueError(message)
    elif scheme == HF_SCHEME:
        if huggingface_hub is None:
            raise ImportError(
                "`from_preset()` requires the `huggingface_hub` package to "
                "load from '{preset}'. "
                "Please install with `pip install huggingface_hub`."
            )
        hf_handle = preset.removeprefix(HF_SCHEME + "://")
        try:
            return huggingface_hub.hf_hub_download(
                repo_id=hf_handle, filename=path
            )
        except HFValidationError as e:
            raise ValueError(
                "Unexpected Hugging Face preset. Hugging Face model handles "
                "should have the form 'hf://{org}/{model}'. For example, "
                f"'hf://username/bert_base_en'. Received: preset={preset}."
            ) from e
        except EntryNotFoundError as e:
            message = str(e)
            if message.find("403 Client Error"):
                raise FileNotFoundError(
                    f"`{path}` doesn't exist in preset directory `{preset}`."
                )
            else:
                raise ValueError(message)
    elif os.path.exists(preset):
        # Assume a local filepath.pyth
        local_path = os.path.join(preset, path)
        if not os.path.exists(local_path):
            raise FileNotFoundError(
                f"`{path}` doesn't exist in preset directory `{preset}`."
            )
        return local_path
    else:
        raise ValueError(
            "Unknown preset identifier. A preset must be a one of:\n"
            "1) a built-in preset identifier like `'bert_base_en'`\n"
            "2) a Kaggle Models handle like "
            "`'kaggle://keras/bert/keras/bert_base_en'`\n"
            "3) a Hugging Face handle like `'hf://username/bert_base_en'`\n"
            "4) a Modelscope handle like `'modelscope://username/bert_base_en'`\n"
            "5) a path to a local preset directory like `'./bert_base_en'`\n"
            "Use `print(cls.presets.keys())` to view all built-in presets for "
            "API symbol `cls`.\n"
            f"Received: preset='{preset}'"
        )


def tf_registered_schemes():
    try:
        import tensorflow as tf

        return tf.io.gfile.get_registered_schemes()
    except ImportError:
        return []


def tf_copy_gfile_to_cache(preset, path):
    """Much of this is adapted from get_file of keras core."""
    if "KERAS_HOME" in os.environ:
        base_dir = os.environ.get("KERAS_HOME")
    else:
        base_dir = os.path.expanduser(os.path.join("~", ".keras"))
    if not os.access(base_dir, os.W_OK):
        base_dir = os.path.join("/tmp", ".keras")

    url = os.path.join(preset, path)
    model_dir = preset.replace("://", "_").replace("-", "_").replace("/", "_")
    local_path = os.path.join(base_dir, "models", model_dir, path)

    if not os.path.exists(local_path):
        print_msg(f"Downloading data from {url}")
        try:
            import tensorflow as tf

            os.makedirs(os.path.dirname(local_path), exist_ok=True)
            tf.io.gfile.copy(url, local_path)
        except Exception as e:
            # gfile.copy will leave an empty file after an error.
            # Work around this bug.
            os.remove(local_path)
            if isinstance(
                e, tf.errors.PermissionDeniedError, tf.errors.NotFoundError
            ):
                raise FileNotFoundError(
                    f"`{path}` doesn't exist in preset directory `{preset}`.",
                ) from e
            raise e

    return local_path


def check_file_exists(preset, path):
    try:
        get_file(preset, path)
    except FileNotFoundError:
        return False
    return True


def _validate_backbone(preset):
    config_path = os.path.join(preset, CONFIG_FILE)
    if not os.path.exists(config_path):
        raise FileNotFoundError(
            f"`{CONFIG_FILE}` is missing from the preset directory `{preset}`."
        )
    try:
        with open(config_path, encoding="utf-8") as config_file:
            json.load(config_file)
    except Exception as e:
        raise ValueError(
            f"Config file `{config_path}` is an invalid json file. "
            f"Error message: {e}"
        )

    weights_path = os.path.join(preset, MODEL_WEIGHTS_FILE)
    sharded_weights_path = os.path.join(preset, "model_*.weights.h5")
    if not os.path.exists(weights_path) and not glob.glob(sharded_weights_path):
        raise FileNotFoundError(
            f"The weights file is missing from the preset directory `{preset}`."
        )


def to_snake_case(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def create_model_card(preset):
    model_card_path = os.path.join(preset, README_FILE)
    markdown_content = ""

    config = load_json(preset, CONFIG_FILE)
    model_name = (
        config["class_name"].replace("Backbone", "")
        if config["class_name"].endswith("Backbone")
        else config["class_name"]
    )

    task_type = None
    if check_file_exists(preset, TASK_CONFIG_FILE):
        task_config = load_json(preset, TASK_CONFIG_FILE)
        task_type = (
            task_config["class_name"].replace(model_name, "")
            if task_config["class_name"].startswith(model_name)
            else task_config["class_name"]
        )

    # YAML
    markdown_content += "---\n"
    markdown_content += "library_name: keras-hub\n"
    if task_type == "CausalLM":
        markdown_content += "pipeline_tag: text-generation\n"
    elif task_type == "TextClassifier":
        markdown_content += "pipeline_tag: text-classification\n"
    markdown_content += "---\n"

    model_link = (
        f"https://keras.io/api/keras_hub/models/{to_snake_case(model_name)}"
    )
    markdown_content += (
        f"This is a [`{model_name}` model]({model_link}) "
        "uploaded using the KerasHub library and can be used with JAX, "
        "TensorFlow, and PyTorch backends.\n"
    )
    if task_type:
        markdown_content += (
            f"This model is related to a `{task_type}` task.\n\n"
        )

    backbone_config = config["config"]
    markdown_content += "Model config:\n"
    for k, v in backbone_config.items():
        markdown_content += f"* **{k}:** {v}\n"
    markdown_content += "\n"
    markdown_content += (
        "This model card has been generated automatically and should be "
        "completed by the model author. See [Model Cards documentation]"
        "(https://huggingface.co/docs/hub/model-cards) for more information.\n"
    )

    with open(model_card_path, "w") as md_file:
        md_file.write(markdown_content)


def delete_model_card(preset):
    model_card_path = os.path.join(preset, README_FILE)
    try:
        os.remove(model_card_path)
    except FileNotFoundError:
        logging.warning(
            f"There was an attempt to delete file `{model_card_path}` but this"
            " file doesn't exist."
        )


@keras_hub_export("keras_hub.upload_preset")
def upload_preset(
    uri,
    preset,
):
    """Upload a preset directory to a model hub.

    Args:
        uri: The URI identifying model to upload to.
             URIs with format
             `kaggle://<KAGGLE_USERNAME>/<MODEL>/<FRAMEWORK>/<VARIATION>`
             will be uploaded to Kaggle Hub while URIs with format
             `hf://[<HF_USERNAME>/]<MODEL>` will be uploaded to the Hugging
             Face Hub.
        preset: The path to the local model preset directory.
    """

    # Check if preset directory exists.
    if not os.path.exists(preset):
        raise FileNotFoundError(f"The preset directory {preset} doesn't exist.")

    _validate_backbone(preset)

    if uri.startswith(KAGGLE_PREFIX):
        if kagglehub is None:
            raise ImportError(
                "Uploading a model to Kaggle Hub requires the `kagglehub` "
                "package. Please install with `pip install kagglehub`."
            )
        kaggle_handle = uri.removeprefix(KAGGLE_PREFIX)
        kagglehub.model_upload(kaggle_handle, preset)
    elif uri.startswith(HF_PREFIX):
        if huggingface_hub is None:
            raise ImportError(
                f"`upload_preset()` requires the `huggingface_hub` package "
                f"to upload to '{uri}'. "
                "Please install with `pip install huggingface_hub`."
            )
        hf_handle = uri.removeprefix(HF_PREFIX)
        try:
            repo_url = huggingface_hub.create_repo(
                repo_id=hf_handle, exist_ok=True
            )
        except HFValidationError as e:
            raise ValueError(
                "Unexpected Hugging Face URI. Hugging Face model handles "
                "should have the form 'hf://[{org}/]{model}'. For example, "
                "'hf://username/bert_base_en' or 'hf://bert_case_en' to "
                f"implicitly upload to your user account. Received: URI={uri}."
            ) from e
        has_model_card = huggingface_hub.file_exists(
            repo_id=repo_url.repo_id, filename=README_FILE
        )
        if not has_model_card:
            # Remote repo doesn't have a model card so a basic model card is
            # automatically generated.
            create_model_card(preset)
        try:
            huggingface_hub.upload_folder(
                repo_id=repo_url.repo_id, folder_path=preset
            )
        finally:
            if not has_model_card:
                # Clean up the preset directory in case user attempts to upload
                # the preset directory into Kaggle hub as well.
                delete_model_card(preset)
    else:
        raise ValueError(
            "Unknown URI. An URI must be a one of:\n"
            "1) a Kaggle Model handle like "
            "`'kaggle://<KAGGLE_USERNAME>/<MODEL>/<FRAMEWORK>/<VARIATION>'`\n"
            "2) a Hugging Face handle like `'hf://[<HF_USERNAME>/]<MODEL>'`\n"
            f"Received: uri='{uri}'."
        )


def load_json(preset, config_file=CONFIG_FILE):
    config_path = get_file(preset, config_file)
    with open(config_path, encoding="utf-8") as config_file:
        config = json.load(config_file)
    return config


def check_config_class(config):
    """Validate a preset is being loaded on the correct class."""
    registered_name = config["registered_name"]
    if registered_name in ("Functional", "Sequential"):
        return keras.Model
    cls = keras.saving.get_registered_object(registered_name)
    if cls is None:
        raise ValueError(
            f"Attempting to load class {registered_name} with "
            "`from_preset()`, but there is no class registered with Keras "
            f"for {registered_name}. Make sure to register any custom "
            "classes with `register_keras_serializable()`."
        )
    return cls


def jax_memory_cleanup(layer):
    # For jax, delete all previous allocated memory to avoid temporarily
    # duplicating variable allocations. torch and tensorflow have stateful
    # variable types and do not need this fix.
    # Skip deletion for sharded arrays to avoid breaking references in
    # distributed setups.
    if keras.config.backend() == "jax":
        for weight in layer.weights:
            if weight._value is not None:
                # Do not delete sharded arrays, as they may be referenced in
                # JAX's distributed computation graph and deletion can cause
                # errors.
                sharding = getattr(weight._value, "sharding", None)
                if sharding is None:
                    weight._value.delete()


def set_dtype_in_config(config, dtype=None):
    if dtype is None:
        return config

    config = config.copy()
    if "dtype" not in config["config"]:
        # Forward `dtype` to the config.
        config["config"]["dtype"] = dtype
    elif (
        "dtype" in config["config"]
        and isinstance(config["config"]["dtype"], dict)
        and "DTypePolicyMap" in config["config"]["dtype"]["class_name"]
    ):
        # If it is `DTypePolicyMap` in `config`, forward `dtype` as its default
        # policy.
        policy_map_config = config["config"]["dtype"]["config"]
        policy_map_config["default_policy"] = dtype
        for k in policy_map_config["policy_map"].keys():
            policy_map_config["policy_map"][k]["config"]["source_name"] = dtype
    return config


def get_preset_loader(preset):
    if not check_file_exists(preset, CONFIG_FILE):
        raise ValueError(
            f"Preset {preset} has no {CONFIG_FILE}. Make sure the URI or "
            "directory you are trying to load is a valid KerasHub preset and "
            "and that you have permissions to read/download from this location."
        )
    # We currently assume all formats we support have a `config.json`, this is
    # true, for Keras, Transformers, and timm. We infer the on disk format by
    # inspecting the `config.json` file.
    config = load_json(preset, CONFIG_FILE)
    if "registered_name" in config:
        # If we see registered_name, we assume a serialized Keras object.
        return KerasPresetLoader(preset, config)
    elif "model_type" in config:
        # Avoid circular import.
        from keras_hub.src.utils.transformers.preset_loader import (
            TransformersPresetLoader,
        )

        # If we see model_type, we assume a Transformers style config.
        return TransformersPresetLoader(preset, config)
    elif "architecture" in config:
        # Avoid circular import.
        from keras_hub.src.utils.timm.preset_loader import TimmPresetLoader

        # If we see "architecture", we assume a timm config. We could make this
        # more robust later on if we need to.
        return TimmPresetLoader(preset, config)

    else:
        contents = json.dumps(config, indent=4)
        raise ValueError(
            f"Unrecognized format for {CONFIG_FILE} in {preset}. "
            "Create a preset with the `save_to_preset` utility on KerasHub "
            f"models. Contents of {CONFIG_FILE}:\n{contents}"
        )


def get_preset_saver(preset):
    # Unlike loading, we only support one form of saving; Keras serialized
    # configs and saved weights. We keep the rough API structure as loading
    # just for simplicity.
    return KerasPresetSaver(preset)


class PresetLoader:
    def __init__(self, preset, config):
        self.config = config
        self.preset = preset

    def get_backbone_kwargs(self, **kwargs):
        backbone_kwargs = {}

        # Forward `dtype` to backbone.
        backbone_kwargs["dtype"] = kwargs.pop("dtype", None)

        # Forward `height` and `width` to backbone when using `TextToImage`.
        if "image_shape" in kwargs:
            backbone_kwargs["image_shape"] = kwargs.pop("image_shape", None)

        return backbone_kwargs, kwargs

    def check_backbone_class(self):
        """Infer the backbone architecture."""
        raise NotImplementedError

    def load_backbone(self, cls, load_weights, **kwargs):
        """Load the backbone model from the preset."""
        raise NotImplementedError

    def load_tokenizer(self, cls, config_file=TOKENIZER_CONFIG_FILE, **kwargs):
        """Load a tokenizer layer from the preset."""
        raise NotImplementedError

    def load_audio_converter(self, cls, **kwargs):
        """Load an audio converter layer from the preset."""
        raise NotImplementedError

    def load_image_converter(self, cls, **kwargs):
        """Load an image converter layer from the preset."""
        raise NotImplementedError

    def load_task(self, cls, load_weights, load_task_weights, **kwargs):
        """Load a task model from the preset.

        By default, we create a task from a backbone and preprocessor with
        default arguments. This means
        """
        if "backbone" not in kwargs:
            backbone_class = cls.backbone_cls
            backbone_kwargs, kwargs = self.get_backbone_kwargs(**kwargs)
            kwargs["backbone"] = self.load_backbone(
                backbone_class, load_weights, **backbone_kwargs
            )
        if "preprocessor" not in kwargs and cls.preprocessor_cls:
            kwargs["preprocessor"] = self.load_preprocessor(
                cls.preprocessor_cls,
            )

        return cls(**kwargs)

    def load_preprocessor(
        self, cls, config_file=PREPROCESSOR_CONFIG_FILE, **kwargs
    ):
        """Load a prepocessor layer from the preset.

        By default, we create a preprocessor from a tokenizer with default
        arguments. This allow us to support transformers checkpoints by
        only converting the backbone and tokenizer.
        """
        kwargs = cls._add_missing_kwargs(self, kwargs)
        return cls(**kwargs)


class KerasPresetLoader(PresetLoader):
    def check_backbone_class(self):
        return check_config_class(self.config)

    def load_backbone(self, cls, load_weights, **kwargs):
        config = self.config.copy()
        backbone_kwargs, kwargs = self.get_backbone_kwargs(**kwargs)
        config["config"] = {**config["config"], **backbone_kwargs}
        backbone = self._load_serialized_object(config, **kwargs)
        if load_weights:
            jax_memory_cleanup(backbone)
            self._load_backbone_weights(backbone)
        return backbone

    def load_tokenizer(self, cls, config_file=TOKENIZER_CONFIG_FILE, **kwargs):
        tokenizer_config = load_json(self.preset, config_file)
        tokenizer = self._load_serialized_object(tokenizer_config, **kwargs)
        if hasattr(tokenizer, "load_preset_assets"):
            tokenizer.load_preset_assets(self.preset)
        return tokenizer

    def load_audio_converter(self, cls, **kwargs):
        converter_config = load_json(self.preset, AUDIO_CONVERTER_CONFIG_FILE)
        return self._load_serialized_object(converter_config, **kwargs)

    def load_image_converter(self, cls, **kwargs):
        converter_config = load_json(self.preset, IMAGE_CONVERTER_CONFIG_FILE)
        return self._load_serialized_object(converter_config, **kwargs)

    def load_task(self, cls, load_weights, load_task_weights, **kwargs):
        # If there is no `task.json` or it's for the wrong class delegate to the
        # super class loader.
        if not check_file_exists(self.preset, TASK_CONFIG_FILE):
            return super().load_task(
                cls, load_weights, load_task_weights, **kwargs
            )
        task_config = load_json(self.preset, TASK_CONFIG_FILE)
        if not issubclass(check_config_class(task_config), cls):
            return super().load_task(
                cls, load_weights, load_task_weights, **kwargs
            )
        # We found a `task.json` with a complete config for our class.
        # Forward backbone args.
        kwargs["dtype"] = self._resolve_dtype(self.config, kwargs)
        backbone_kwargs, kwargs = self.get_backbone_kwargs(**kwargs)
        if "backbone" in task_config["config"]:
            backbone_config = task_config["config"]["backbone"]["config"]
            backbone_config = {**backbone_config, **backbone_kwargs}
            task_config["config"]["backbone"]["config"] = backbone_config
        task = self._load_serialized_object(task_config, **kwargs)
        if task.preprocessor and hasattr(
            task.preprocessor, "load_preset_assets"
        ):
            task.preprocessor.load_preset_assets(self.preset)
        if load_weights:
            has_task_weights = check_file_exists(self.preset, TASK_WEIGHTS_FILE)
            if has_task_weights and load_task_weights:
                jax_memory_cleanup(task)
                task_weights = get_file(self.preset, TASK_WEIGHTS_FILE)
                task.load_task_weights(task_weights)
            else:
                jax_memory_cleanup(task.backbone)
            self._load_backbone_weights(task.backbone)
        return task

    def _resolve_dtype(self, config, kwargs):
        """Resolves the Model's dtype based on the provided config and kwargs.

        The data type is resolved based on the following priority:
        1. If a user specified dtype is passed, use that.
        2. If no user specified dtype is passed, and the save dtype is castable
          to the current keras default dtype convert weights on load (float type
          to float type).
        3. If not user specified dtype is passed, and the save dtype is not
          castable to the current default dtype (quantized dtypes). Load the
          saved types verbatim.

        Args:
            config: dict. The model configuration.
            kwargs: dict. Additional keyword arguments, potentially including
             `dtype`.

        Returns:
            str, dict, or DTypePolicy. The resolved dtype.
        """
        # 1. If a user specified dtype is passed, use that.
        if "dtype" in kwargs and kwargs["dtype"] is not None:
            return kwargs["dtype"]

        saved_dtype = config.get("config", {}).get("dtype")

        # If there's no saved dtype, we don't need to do anything.
        if saved_dtype is None:
            return None

        # 2. Check whether the saved dtype is a simple float type.
        policy_name = saved_dtype.get("config", {}).get("name")
        if policy_name and tensor_utils.is_float_dtype(policy_name):
            # If the saved dtype is a float, we can safely cast to the default
            # backend float type.
            if policy_name != keras.config.dtype_policy().name:
                logging.info(
                    f"Converting weights saved as {policy_name} "
                    "to the current Keras dtype policy "
                    f"{keras.config.dtype_policy()}"
                )
            return keras.config.dtype_policy()
        else:
            # 3. Otherwise, the dtype is a complex object (e.g. a
            # DTypePolicyMap for quantization), and should be used as is.
            return saved_dtype

    def load_preprocessor(
        self, cls, config_file=PREPROCESSOR_CONFIG_FILE, **kwargs
    ):
        # If there is no `preprocessing.json` or it's for the wrong class,
        # delegate to the super class loader.
        if not check_file_exists(self.preset, config_file):
            return super().load_preprocessor(cls, **kwargs)
        preprocessor_json = load_json(self.preset, config_file)
        if not issubclass(check_config_class(preprocessor_json), cls):
            return super().load_preprocessor(cls, **kwargs)
        # We found a `preprocessing.json` with a complete config for our class.
        preprocessor = self._load_serialized_object(preprocessor_json, **kwargs)
        if hasattr(preprocessor, "load_preset_assets"):
            preprocessor.load_preset_assets(self.preset)
        return preprocessor

    def _load_serialized_object(self, config, **kwargs):
        # `dtype` in config might be a serialized `DTypePolicy` or
        # `DTypePolicyMap`. Ensure that `dtype` is properly configured.
        dtype = kwargs.pop("dtype", None)
        config = set_dtype_in_config(config, dtype)

        config["config"] = {**config["config"], **kwargs}
        return keras.saving.deserialize_keras_object(config)

    def _get_sharded_filenames(self, config_path):
        with open(config_path, encoding="utf-8") as config_file:
            config = json.load(config_file)
        weight_map = config["weight_map"]
        filenames = set()
        for v in weight_map.values():
            if isinstance(v, list):
                filenames.update(v)
            else:
                filenames.add(v)
        return sorted(filenames)

    def _load_backbone_weights(self, backbone):
        # Detect if the backbone is sharded or not.
        has_single_file_weights = check_file_exists(
            self.preset, MODEL_WEIGHTS_FILE
        )
        if has_single_file_weights:
            filepath = get_file(self.preset, MODEL_WEIGHTS_FILE)
        else:
            if not sharded_weights_available():
                raise RuntimeError(
                    "Sharded weights loading is not supported in the current "
                    f"Keras version {keras.__version__}. "
                    "Please update to a newer version."
                )
            filepath = get_file(self.preset, SHARDED_MODEL_WEIGHTS_CONFIG_FILE)
            sharded_filenames = self._get_sharded_filenames(filepath)
            for sharded_filename in sharded_filenames:
                # Download the sharded weights.
                _ = get_file(self.preset, sharded_filename)
        backbone.load_weights(filepath)


class KerasPresetSaver:
    def __init__(self, preset_dir):
        os.makedirs(preset_dir, exist_ok=True)
        self.preset_dir = preset_dir

    def save_backbone(self, backbone, max_shard_size=10):
        self._save_serialized_object(backbone, config_file=CONFIG_FILE)
        self._save_metadata(backbone)

        # Save the weights.
        backbone_size_in_bytes = self._get_variables_size_in_bytes(
            backbone.variables
        )
        backbone_size_in_gb = backbone_size_in_bytes / (1024**3)
        # If the size of the backbone is larger than `max_shard_size`, save
        # sharded weights.
        if (
            sharded_weights_available()
            and max_shard_size is not None
            and backbone_size_in_gb > max_shard_size
        ):
            backbone_sharded_weights_config_path = os.path.join(
                self.preset_dir, SHARDED_MODEL_WEIGHTS_CONFIG_FILE
            )
            backbone.save_weights(
                backbone_sharded_weights_config_path,
                max_shard_size=max_shard_size,
            )
        else:
            backbone_weight_path = os.path.join(
                self.preset_dir, MODEL_WEIGHTS_FILE
            )
            backbone.save_weights(backbone_weight_path)

    def save_tokenizer(self, tokenizer):
        config_file = TOKENIZER_CONFIG_FILE
        if hasattr(tokenizer, "config_file"):
            config_file = tokenizer.config_file
        self._save_serialized_object(tokenizer, config_file)
        # Save assets.
        subdir = config_file.split(".")[0]
        asset_dir = os.path.join(self.preset_dir, ASSET_DIR, subdir)
        os.makedirs(asset_dir, exist_ok=True)
        tokenizer.save_assets(asset_dir)

    def save_audio_converter(self, converter):
        self._save_serialized_object(converter, AUDIO_CONVERTER_CONFIG_FILE)

    def save_image_converter(self, converter):
        self._save_serialized_object(converter, IMAGE_CONVERTER_CONFIG_FILE)

    def save_task(self, task, max_shard_size=10):
        # Save task specific config and weights.
        self._save_serialized_object(task, TASK_CONFIG_FILE)
        if task.has_task_weights():
            task_weight_path = os.path.join(self.preset_dir, TASK_WEIGHTS_FILE)
            task.save_task_weights(task_weight_path)
        # Save backbone.
        if hasattr(task.backbone, "save_to_preset"):
            task.backbone.save_to_preset(
                self.preset_dir, max_shard_size=max_shard_size
            )
        else:
            # Allow saving a `keras.Model` that is not a backbone subclass.
            self.save_backbone(task.backbone, max_shard_size=max_shard_size)
        # Save preprocessor.
        if task.preprocessor and hasattr(task.preprocessor, "save_to_preset"):
            task.preprocessor.save_to_preset(self.preset_dir)
        else:
            # Allow saving a `keras.Layer` that is not a preprocessor subclass.
            self.save_preprocessor(task.preprocessor)

    def save_preprocessor(self, preprocessor):
        config_file = PREPROCESSOR_CONFIG_FILE
        if hasattr(preprocessor, "config_file"):
            config_file = preprocessor.config_file
        self._save_serialized_object(preprocessor, config_file)
        for layer in preprocessor._flatten_layers(include_self=False):
            if hasattr(layer, "save_to_preset"):
                layer.save_to_preset(self.preset_dir)

    def _recursive_pop(self, config, key):
        """Remove a key from a nested config object"""
        config.pop(key, None)
        for value in config.values():
            if isinstance(value, dict):
                self._recursive_pop(value, key)

    def _save_serialized_object(self, layer, config_file):
        config_path = os.path.join(self.preset_dir, config_file)
        config = keras.saving.serialize_keras_object(layer)
        config_to_skip = ["compile_config", "build_config"]
        for key in config_to_skip:
            self._recursive_pop(config, key)
        with open(config_path, "w") as config_file:
            config_file.write(json.dumps(config, indent=4))

    def _save_metadata(self, layer):
        from keras_hub.src.models.task import Task
        from keras_hub.src.version import __version__ as keras_hub_version

        # Find all tasks that are compatible with the backbone.
        # E.g. for `BertBackbone` we would have `TextClassifier` and `MaskedLM`.
        # For `ResNetBackbone` we would have `ImageClassifier`.
        tasks = list_subclasses(Task)
        tasks = filter(lambda x: x.backbone_cls is type(layer), tasks)
        tasks = [task.__base__.__name__ for task in tasks]
        # Keep task list alphabetical.
        tasks = sorted(tasks)

        keras_version = keras.version() if hasattr(keras, "version") else None
        metadata = {
            "keras_version": keras_version,
            "keras_hub_version": keras_hub_version,
            "parameter_count": layer.count_params(),
            "date_saved": datetime.datetime.now().strftime("%Y-%m-%d@%H:%M:%S"),
            "tasks": tasks,
        }
        metadata_path = os.path.join(self.preset_dir, METADATA_FILE)
        with open(metadata_path, "w") as metadata_file:
            metadata_file.write(json.dumps(metadata, indent=4))

    def _get_variables_size_in_bytes(self, variables):
        unique_variables = {}
        for v in variables:
            if id(v) not in unique_variables:
                unique_variables[id(v)] = (v.shape, v.dtype)
        total_memory_size = 0
        for shape, dtype in unique_variables.values():
            total_memory_size += get_tensor_size_in_bits(shape, dtype)
        return total_memory_size / 8



import os
import re
from typing import Iterable

import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.tokenizers import tokenizer
from keras_hub.src.utils.tensor_utils import convert_to_ragged_batch
from keras_hub.src.utils.tensor_utils import is_int_dtype
from keras_hub.src.utils.tensor_utils import is_string_dtype
from keras_hub.src.utils.tensor_utils import preprocessing_function

try:
    import tensorflow as tf
    import tensorflow_text as tf_text
except ImportError:
    tf = None
    tf_text = None

PARSEQ_VOCAB = list(
    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"
    "\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
)

VOCAB_FILENAME = "vocabulary.txt"


@keras_hub_export(
    [
        "keras_hub.tokenizers.PARSeqTokenizer",
        "keras_hub.models.PARSeqTokenizer",
    ]
)
class PARSeqTokenizer(tokenizer.Tokenizer):
    """A Tokenizer for PARSeq models, designed for OCR tasks.

    This tokenizer converts strings into sequences of integer IDs or string
    tokens, and vice-versa. It supports various preprocessing steps such as
    whitespace removal, Unicode normalization, and limiting the maximum label
    length. It also provides functionality to save and load the vocabulary
    from a file.

    Args:
        vocabulary: str. A string or iterable representing the vocabulary to
            use. If a string, it's treated as the path to a vocabulary file.
            If an iterable, it's treated as a list of characters forming
            the vocabulary. Defaults to `PARSEQ_VOCAB`.
        remove_whitespace: bool. Whether to remove whitespace characters from
            the input. Defaults to `True`.
        normalize_unicode: bool. Whether to normalize Unicode characters in the
            input using NFKD normalization and remove non-ASCII characters.
            Defaults to `True`.
        max_label_length: int. The maximum length of the tokenized output.
            Longer labels will be truncated. Defaults to `25`.
        dtype: str. The data type of the tokenized output. Must be an integer
            type (e.g., "int32") or a string type ("string").
            Defaults to `"int32"`.
        **kwargs: Additional keyword arguments passed to the base
            `keras.layers.Layer` constructor.
    """

    def __init__(
        self,
        vocabulary=PARSEQ_VOCAB,
        remove_whitespace=True,
        normalize_unicode=True,
        max_label_length=25,
        dtype="int32",
        **kwargs,
    ):
        if not is_int_dtype(dtype) and not is_string_dtype(dtype):
            raise ValueError(
                "Output dtype must be an integer type or a string. "
                f"Received: dtype={dtype}"
            )
        super().__init__(dtype=dtype, **kwargs)
        self.remove_whitespace = remove_whitespace
        self.normalize_unicode = normalize_unicode
        self.max_label_length = max_label_length
        self.file_assets = [VOCAB_FILENAME]

        self.set_vocabulary(vocabulary)

    def save_assets(self, dir_path):
        path = os.path.join(dir_path, VOCAB_FILENAME)
        with open(path, "w", encoding="utf-8") as file:
            for token in self.vocabulary:
                file.write(f"{token}\n")

    def load_assets(self, dir_path):
        path = os.path.join(dir_path, VOCAB_FILENAME)
        self.set_vocabulary(path)

    def set_vocabulary(self, vocabulary):
        """Set the tokenizer vocabulary to a file or list of strings."""
        if vocabulary is None:
            self.vocabulary = None
            return

        if isinstance(vocabulary, str):
            with open(vocabulary, "r", encoding="utf-8") as file:
                self.vocabulary = [line.rstrip() for line in file]
                self.vocabulary = "".join(self.vocabulary)
        elif isinstance(vocabulary, Iterable):
            self.vocabulary = "".join(vocabulary)
        else:
            raise ValueError(
                "Vocabulary must be an file path or list of terms. "
                f"Received: vocabulary={vocabulary}"
            )

        self.lowercase_only = self.vocabulary == self.vocabulary.lower()
        self.uppercase_only = self.vocabulary == self.vocabulary.upper()
        escaped_charset = re.escape(self.vocabulary)  # Escape for safe regex
        self.unsupported_regex = f"[^{escaped_charset}]"
        self._itos = ("[E]",) + tuple(self.vocabulary) + ("[B]", "[P]")
        self._stoi = {s: i for i, s in enumerate(self._itos)}

        self._add_special_token("[B]", "start_token")
        self._add_special_token("[E]", "end_token")
        self._add_special_token("[P]", "pad_token")
        # Create lookup tables.
        self.char_to_id = tf.lookup.StaticHashTable(
            initializer=tf.lookup.KeyValueTensorInitializer(
                keys=list(self._stoi.keys()),
                values=list(self._stoi.values()),
                key_dtype=tf.string,
                value_dtype=tf.int32,
            ),
            default_value=self._stoi["[E]"],
        )
        self.id_to_char = tf.lookup.StaticHashTable(
            initializer=tf.lookup.KeyValueTensorInitializer(
                keys=list(self._stoi.values()),
                values=list(self._stoi.keys()),
                key_dtype=tf.int32,
                value_dtype=tf.string,
            ),
            default_value=self.pad_token,
        )

    def get_vocabulary(self):
        """Get the tokenizer vocabulary as a list of strings tokens."""
        return list(self.vocabulary)

    def id_to_token(self, id):
        if id >= self.vocabulary_size() or id < 0:
            raise ValueError(
                f"`id` must be in range [0, {self.vocabulary_size() - 1}]. "
                f"Received: {id}"
            )
        return self._itos[id]

    def token_to_id(self, token):
        return self._stoi[token]

    def _preprocess(self, inputs):
        """Performs preprocessing include only characters from ASCII."""
        if self.remove_whitespace:
            inputs = tf.strings.regex_replace(inputs, r"\s+", "")

        if self.normalize_unicode:
            inputs = tf_text.normalize_utf8(inputs, normalization_form="NFKD")
            inputs = tf.strings.regex_replace(inputs, r"[^!-~]", "")

        if self.lowercase_only:
            inputs = tf.strings.lower(inputs)
        elif self.uppercase_only:
            inputs = tf.strings.upper(inputs)

        inputs = tf.strings.regex_replace(inputs, self.unsupported_regex, "")
        inputs = tf.strings.substr(inputs, 0, self.max_label_length)

        return inputs

    @preprocessing_function
    def tokenize(self, inputs):
        inputs = tf.convert_to_tensor(inputs)
        unbatched = inputs.shape.rank == 0
        if unbatched:
            inputs = tf.expand_dims(inputs, 0)

        inputs = tf.map_fn(
            self._preprocess, inputs, fn_output_signature=tf.string
        )

        token_ids = tf.cond(
            tf.size(inputs) > 0,
            lambda: self.char_to_id.lookup(
                tf.strings.unicode_split(inputs, "UTF-8")
            ),
            lambda: tf.RaggedTensor.from_row_splits(
                values=tf.constant([], dtype=tf.int32),
                row_splits=tf.constant([0], dtype=tf.int64),
            ),
        )
        if unbatched:
            token_ids = tf.squeeze(token_ids, 0)
            tf.ensure_shape(token_ids, shape=[self.max_label_length])
        return token_ids

    @preprocessing_function
    def detokenize(self, inputs):
        inputs, unbatched, rectangular = convert_to_ragged_batch(inputs)
        # tf-text sentencepiece does not handle int64.
        inputs = tf.cast(inputs, "int32")
        outputs = self.id_to_char.lookup(inputs)
        if unbatched:
            outputs = tf.squeeze(outputs, 0)
        return outputs

    def vocabulary_size(self):
        """Get the integer size of the tokenizer vocabulary."""
        return len(self.vocabulary) + 3

    def compute_output_spec(self, input_spec):
        return keras.KerasTensor(
            input_spec.shape + (self.max_label_length,),
            dtype=self.compute_dtype,
        )



import math

from keras import layers
from keras import ops

from keras_hub.src.utils.keras_utils import standardize_data_format


def quick_gelu(x):
    return x * ops.sigmoid(1.702 * x)


class CLIPVisionEmbedding(layers.Layer):
    def __init__(
        self,
        hidden_dim,
        patch_size,
        image_size,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.hidden_dim = int(hidden_dim)
        self.patch_size = int(patch_size)
        self.image_size = int(image_size)
        data_format = standardize_data_format(data_format)
        self.data_format = data_format
        num_patches = (image_size // patch_size) ** 2
        self.num_positions = num_patches + 1

        self.patch_embedding = layers.Conv2D(
            hidden_dim,
            kernel_size=patch_size,
            strides=patch_size,
            data_format=data_format,
            use_bias=False,
            dtype=dtype,
            name="patch_embedding",
        )
        self.position_embedding = layers.Embedding(
            num_patches + 1, hidden_dim, dtype=dtype, name="position_embedding"
        )

    def build(self, input_shape):
        self.class_embedding = self.add_weight(
            shape=(self.hidden_dim,),
            initializer="random_normal",
            dtype=self.variable_dtype,
            name="class_embedding",
        )
        self.position_ids = self.add_weight(
            shape=(1, self.num_positions),
            initializer="zeros",
            # Let the backend determine the int dtype. For example, tf
            # requires int64 for correct device placement, whereas jax and torch
            # don't.
            dtype=int,
            trainable=False,
            name="position_ids",
        )
        self.patch_embedding.build(input_shape)
        self.position_embedding.build(self.position_ids.shape)

    def call(self, inputs, training=None):
        x = inputs
        batch_size = ops.shape(x)[0]
        patch_embeddings = self.patch_embedding(x, training=training)
        if self.data_format == "channels_last":
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, -1, self.hidden_dim)
            )
        else:
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, self.hidden_dim, -1)
            )
            patch_embeddings = ops.transpose(patch_embeddings, (0, 2, 1))
        class_embeddings = ops.expand_dims(self.class_embedding, axis=(0, 1))
        class_embeddings = ops.tile(class_embeddings, (batch_size, 1, 1))
        position_embeddings = self.position_embedding(self.position_ids)
        embeddings = ops.concatenate(
            [class_embeddings, patch_embeddings], axis=1
        )
        return ops.add(embeddings, position_embeddings)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "patch_size": self.patch_size,
                "image_size": self.image_size,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        output_shape = [input_shape[0], None, self.hidden_dim]
        if self.data_format == "channels_last":
            if input_shape[1] is not None and input_shape[2] is not None:
                patch_num = input_shape[1] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        else:
            if input_shape[2] is not None and input_shape[3] is not None:
                patch_num = input_shape[2] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        return output_shape


class CLIPEncoderLayer(layers.Layer):
    def __init__(
        self,
        hidden_dim,
        num_heads,
        intermediate_dim,
        intermediate_activation="quick_gelu",
        use_causal_mask=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if hidden_dim % num_heads != 0:
            raise ValueError(
                "`hidden_dim` must be divisible by `num_heads`. "
                f"Received: hidden_dim={hidden_dim}, num_heads={num_heads}"
            )
        self.hidden_dim = hidden_dim
        self.num_heads = num_heads
        self.intermediate_dim = intermediate_dim
        self.intermediate_activation = intermediate_activation
        self.use_causal_mask = use_causal_mask

        if intermediate_activation == "quick_gelu":
            intermediate_activation = quick_gelu

        self.layer_norm_1 = layers.LayerNormalization(
            epsilon=1e-5, dtype=self.dtype_policy, name="layer_norm_1"
        )
        self.attention = layers.MultiHeadAttention(
            num_heads,
            hidden_dim // num_heads,
            dtype=self.dtype_policy,
            name="attention",
        )
        self.layer_norm_2 = layers.LayerNormalization(
            epsilon=1e-5, dtype=self.dtype_policy, name="layer_norm_2"
        )
        self.dense_1 = layers.Dense(
            self.intermediate_dim, dtype=self.dtype_policy, name="dense_1"
        )
        self.activation = layers.Activation(
            intermediate_activation, dtype=self.dtype_policy, name="activation"
        )
        self.dense_2 = layers.Dense(
            self.hidden_dim, dtype=self.dtype_policy, name="dense_2"
        )

    def build(self, input_shape):
        self.layer_norm_1.build(input_shape)
        self.attention.build(input_shape, input_shape, input_shape)
        self.layer_norm_2.build(input_shape)
        self.dense_1.build(input_shape)
        input_shape = self.dense_1.compute_output_shape(input_shape)
        self.dense_2.build(input_shape)

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        outputs_shape[-1] = self.hidden_dim
        return outputs_shape

    def call(self, x, training=None):
        residual = x
        x = self.layer_norm_1(x)
        x = self.attention(
            x, x, x, training=training, use_causal_mask=self.use_causal_mask
        )
        x = ops.add(residual, x)

        residual = x
        x = self.dense_1(self.layer_norm_2(residual))
        x = self.activation(x)
        x = self.dense_2(x)
        x = ops.add(residual, x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "num_heads": self.num_heads,
                "intermediate_dim": self.intermediate_dim,
                "intermediate_activation": self.intermediate_activation,
                "use_causal_mask": self.use_causal_mask,
            }
        )
        return config


class CLIPVisionPooler(layers.Layer):
    """The vision pooler layer of CLIP.

    `CLIPVisionPooler` will extracts the first token (index `0`) from the
    sequence of the vision embeddings as the pooled outputs.

    Call arguments:
        vision_embeddings: A tensor of shape
            `(batch_size, sequence_length, hidden_dim)`.
    """

    def call(self, vision_embeddings):
        return vision_embeddings[:, 0, :]

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])


class CLIPTextPooler(layers.Layer):
    """The text pooler layer of CLIP.

    `CLIPTextPooler` extracts the text embeddings at the positions of EOS tokens
    as the pooled outputs.

    Call arguments:
        text_embeddings: A tensor of shape
            `(batch_size, sequence_length, hidden_dim)`.
        token_ids: A tensor of shape `(batch_size, max_tokens)`, used to
            identify the positions of EOS tokens.
    """

    def call(self, text_embeddings, token_ids):
        # `keepdims` is not supported in `keras<=3.1`.
        eos_index = ops.argmax(token_ids, axis=-1)
        eos_index = ops.expand_dims(eos_index, axis=-1)
        eos_index = ops.expand_dims(eos_index, axis=-1)
        pooled_outputs = ops.take_along_axis(text_embeddings, eos_index, axis=1)
        return ops.squeeze(pooled_outputs, axis=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])


class CLIPHead(layers.Layer):
    """The head layer of CLIP.

    `CLIPHead` takes `vision_embedding` and `text_embedding` as inputs to
    compute the corresponding logits. Both embeddings are L2 normalized and used
    to compute pairwise cosine similarity. The resulting logits are then scaled
    by a learnable `logit_scale` parameter.

    Call arguments:
        vision_embedding: A tensor of shape `(batch_size, hidden_dim)`.
        text_embedding: A tensor of shape `(batch_size, hidden_dim)`.
    """

    def build(self, input_shape):
        self.logit_scale = self.add_weight(
            shape=(),
            initializer=lambda *a, **kw: math.log(1 / 0.07),
            trainable=True,
            dtype=self.variable_dtype,
            name="logit_scale",
        )

    def call(self, vision_embedding, text_embedding):
        normalized_vision_embedding = ops.sqrt(
            ops.sum(ops.power(vision_embedding, 2), axis=-1, keepdims=True)
        )
        normalized_text_embedding = ops.sqrt(
            ops.sum(ops.power(text_embedding, 2), axis=-1, keepdims=True)
        )
        vision_embedding = vision_embedding / normalized_vision_embedding
        text_embedding = text_embedding / normalized_text_embedding
        logit_scale = ops.exp(self.logit_scale)
        text_logits = (
            ops.matmul(
                text_embedding,
                ops.transpose(vision_embedding),
            )
            * logit_scale
        )
        vision_logits = ops.transpose(text_logits)
        return vision_logits, text_logits

    def compute_output_shape(
        self, vision_embedding_shape, text_embedding_shape
    ):
        vision_logits_shape = (
            vision_embedding_shape[0],
            text_embedding_shape[0],
        )
        text_logits_shape = (
            text_embedding_shape[0],
            vision_embedding_shape[0],
        )
        return vision_logits_shape, text_logits_shape



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.task import Task


@keras_hub_export(
    [
        "keras_hub.models.ObjectDetector",
        "keras_hub.models.ImageObjectDetector",
    ]
)
class ObjectDetector(Task):
    """Base class for all image object detection tasks.

    The `ObjectDetector` tasks wrap a `keras_hub.models.Backbone` and
    a `keras_hub.models.Preprocessor` to create a model that can be used for
    object detection. `ObjectDetector` tasks take an additional
    `num_classes` argument, controlling the number of predicted output classes.

    To fine-tune with `fit()`, pass a dataset containing tuples of `(x, y)`
    labels where `x` is a string and `y` is dictionary with `boxes` and
    `classes`.

    All `ObjectDetector` tasks include a `from_preset()` constructor which
    can be used to load a pre-trained config and weights.
    """

    def compile(
        self,
        optimizer="auto",
        box_loss="auto",
        classification_loss="auto",
        metrics=None,
        **kwargs,
    ):
        """Configures the `ObjectDetector` task for training.

        The `ObjectDetector` task extends the default compilation signature
        of `keras.Model.compile` with defaults for `optimizer`, `loss`, and
        `metrics`. To override these defaults, pass any value
        to these arguments during compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default optimizer
                for the given model and task. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer` values.
            box_loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, where a
                `keras.losses.Huber` loss will be
                applied for the object detector task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            classification_loss: `"auto"`, a loss name, or a `keras.losses.Loss`
                instance. Defaults to `"auto"`, where a
                `keras.losses.BinaryFocalCrossentropy` loss will be
                applied for the object detector task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `a list of metrics to be evaluated by
                the model during training and testing. Defaults to `None`.
                See `keras.Model.compile` and `keras.metrics` for
                more info on possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if optimizer == "auto":
            optimizer = keras.optimizers.Adam(5e-5)
        if box_loss == "auto":
            box_loss = keras.losses.Huber(reduction="sum")
        if classification_loss == "auto":
            activation = getattr(self, "activation", None)
            activation = keras.activations.get(activation)
            from_logits = activation != keras.activations.sigmoid
            classification_loss = keras.losses.BinaryFocalCrossentropy(
                from_logits=from_logits, reduction="sum"
            )
        if metrics is not None:
            raise ValueError("User metrics not yet supported")

        losses = {
            "bbox_regression": box_loss,
            "cls_logits": classification_loss,
        }

        super().compile(
            optimizer=optimizer,
            loss=losses,
            metrics=metrics,
            **kwargs,
        )



"""Qwen preset configurations."""

backbone_presets = {
    "qwen2.5_0.5b_en": {
        "metadata": {
            "description": ("24-layer Qwen model with 0.5 billion parameters."),
            "params": 494032768,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_0.5b_en/1",
    },
    "qwen2.5_3b_en": {
        "metadata": {
            "description": ("36-layer Qwen model with 3.1 billion parameters."),
            "params": 3085938688,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_3b_en/1",
    },
    "qwen2.5_7b_en": {
        "metadata": {
            "description": ("48-layer Qwen model with 7 billion parameters."),
            "params": 6993420288,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_7b_en/3",
    },
    "qwen2.5_instruct_0.5b_en": {
        "metadata": {
            "description": (
                "Instruction fine-tuned 24-layer Qwen model with 0.5 "
                "billion parameters."
            ),
            "params": 494032768,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_instruct_0.5b_en/1",
    },
    "qwen2.5_instruct_32b_en": {
        "metadata": {
            "description": (
                "Instruction fine-tuned 64-layer Qwen model with 32 "
                "billion parameters."
            ),
            "params": 32763876352,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_instruct_32b_en/2",
    },
    "qwen2.5_instruct_72b_en": {
        "metadata": {
            "description": (
                "Instruction fine-tuned 80-layer Qwen model with 72 "
                "billion parameters."
            ),
            "params": 72706203648,
            "path": "qwen",
        },
        "kaggle_handle": "kaggle://keras/qwen/keras/qwen2.5_instruct_72b_en/2",
    },
}



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.preprocessor import Preprocessor
from keras_hub.src.utils.tensor_utils import preprocessing_function


@keras_hub_export("keras_hub.models.DepthEstimatorPreprocessor")
class DepthEstimatorPreprocessor(Preprocessor):
    """Base class for depth estimation preprocessing layers.

    `DepthEstimatorPreprocessor` tasks wraps a
    `keras_hub.layers.ImageConverter` to create a preprocessing layer for
    depth estimation tasks. It is intended to be paired with a
    `keras_hub.models.DepthEstimator` task.

    All `DepthEstimatorPreprocessor` take inputs three inputs, `x`, `y`, and
    `sample_weight`. `x`, the first input, should always be included. It can
    be a image or batch of images. See examples below. `y` and `sample_weight`
    are optional inputs that will be passed through unaltered. Usually, `y` will
    be the depths, and `sample_weight` will not be provided.

    The layer will output either `x`, an `(x, y)` tuple if depths were provided,
    or an `(x, y, sample_weight)` tuple if depths and sample weight were
    provided. `x` will be the input images after all model preprocessing has
    been applied.

    All `DepthEstimatorPreprocessor` tasks include a `from_preset()`
    constructor which can be used to load a pre-trained config.
    You can call the `from_preset()` constructor directly on this base class, in
    which case the correct class for your model will be automatically
    instantiated.

    Examples.
    ```python
    preprocessor = keras_hub.models.DepthEstimatorPreprocessor.from_preset(
        "depth_anything_v2_small",
    )

    # Resize a single image for DepthAnythingV2 Small.
    x = np.random.randint(0, 256, (512, 512, 3))
    x = preprocessor(x)

    # Resize a labeled image.
    x = np.random.randint(0, 256, (512, 512, 3))
    y = np.random.uniform(0, 10, size=(512, 512))
    x, y = preprocessor(x, y)

    # Resize a batch of labeled images.
    x = [
        np.random.randint(0, 256, (512, 512, 3)),
        np.zeros((512, 512, 3)),
    ]
    y = [
        np.random.uniform(0, 10, size=(512, 512)),
        np.random.uniform(0, 10, size=(512, 512)),
    ]
    x, y = preprocessor(x, y)

    # Use a `tf.data.Dataset`.
    ds = tf.data.Dataset.from_tensor_slices((x, y)).batch(2)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)
    ```
    """

    def __init__(
        self,
        image_converter=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.image_converter = image_converter

    @preprocessing_function
    def call(self, x, y=None, sample_weight=None):
        if self.image_converter:
            x = self.image_converter(x)
        return keras.utils.pack_x_y_sample_weight(x, y, sample_weight)



import math

import keras

from keras_hub.src.models.d_fine.d_fine_utils import (
    multi_scale_deformable_attention_v2,
)


class DFineMultiscaleDeformableAttention(keras.layers.Layer):
    """Multi-scale deformable attention layer for D-FINE models.

    This layer implements the multi-scale deformable attention mechanism, which
    is the core of the cross-attention in each `DFineDecoderLayer`. It allows
    the model to attend to a small set of key sampling points around a reference
    point across multiple feature levels from the encoder.

    The layer computes sampling locations and attention weights based on the
    input queries, enabling the model to focus on relevant features across
    multiple feature levels and spatial positions.

    Args:
        hidden_dim: int, Hidden dimension size for the attention mechanism.
        decoder_attention_heads: int, Number of attention heads.
        num_feature_levels: int, Number of feature levels to attend to.
        decoder_offset_scale: float, Scaling factor for sampling offsets.
        decoder_method: str, Method used for deformable attention computation.
        decoder_n_points: int or list, Number of sampling points per level.
            If int, the same number of points is used for all levels.
            If list, specifies points for each level individually.
        num_queries: int, Number of queries in the attention mechanism.
        spatial_shapes: list, List of spatial shapes for different
            feature levels.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        hidden_dim,
        decoder_attention_heads,
        num_feature_levels,
        decoder_offset_scale,
        decoder_method,
        decoder_n_points,
        num_queries,
        spatial_shapes,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.hidden_dim = hidden_dim
        self.num_queries = num_queries
        self.n_heads = decoder_attention_heads
        self.n_levels = num_feature_levels
        self.offset_scale = decoder_offset_scale
        self.decoder_method = decoder_method
        self.decoder_n_points = decoder_n_points
        self.spatial_shapes = spatial_shapes
        if isinstance(self.decoder_n_points, list):
            self.num_points = self.decoder_n_points
        else:
            self.num_points = [
                self.decoder_n_points for _ in range(self.n_levels)
            ]
        self._num_points_scale = [
            1.0 / n_points_at_level
            for n_points_at_level in self.num_points
            for _ in range(n_points_at_level)
        ]
        self.total_points = self.n_heads * sum(self.num_points)
        self.ms_deformable_attn_core = multi_scale_deformable_attention_v2

    def build(self, input_shape):
        sampling_offsets_output_shape = (
            input_shape[1],
            self.n_heads,
            sum(self.num_points),
            2,
        )
        self.sampling_offsets = keras.layers.EinsumDense(
            "abc,cdef->abdef",
            output_shape=sampling_offsets_output_shape,
            bias_axes="def",
            kernel_initializer="zeros",
            bias_initializer="zeros",
            name="sampling_offsets",
            dtype=self.dtype_policy,
        )
        self.sampling_offsets.build(input_shape)
        attention_weights_output_shape = (
            input_shape[1],
            self.n_heads,
            sum(self.num_points),
        )
        self.attention_weights = keras.layers.EinsumDense(
            "abc,cde->abde",
            output_shape=attention_weights_output_shape,
            bias_axes="de",
            kernel_initializer="zeros",
            bias_initializer="zeros",
            name="attention_weights",
            dtype=self.dtype_policy,
        )
        self.attention_weights.build(input_shape)
        if self.sampling_offsets.bias is not None:
            thetas = keras.ops.arange(
                self.n_heads, dtype=self.variable_dtype
            ) * (2.0 * math.pi / self.n_heads)
            grid_init = keras.ops.stack(
                [keras.ops.cos(thetas), keras.ops.sin(thetas)], axis=-1
            )
            grid_init = grid_init / keras.ops.max(
                keras.ops.abs(grid_init), axis=-1, keepdims=True
            )
            grid_init = keras.ops.reshape(grid_init, (self.n_heads, 1, 2))
            grid_init = keras.ops.tile(grid_init, [1, sum(self.num_points), 1])
            scaling = []
            for n in self.num_points:
                scaling.append(
                    keras.ops.arange(1, n + 1, dtype=self.variable_dtype)
                )
            scaling = keras.ops.concatenate(scaling, axis=0)
            scaling = keras.ops.reshape(scaling, (1, -1, 1))
            grid_init *= scaling
            self.sampling_offsets.bias.assign(grid_init)
        self.num_points_scale = self.add_weight(
            name="num_points_scale",
            shape=(len(self._num_points_scale),),
            initializer=keras.initializers.Constant(self._num_points_scale),
            trainable=False,
        )
        super().build(input_shape)

    def compute_attention(
        self, hidden_states, reference_points, spatial_shapes
    ):
        batch_size = keras.ops.shape(hidden_states)[0]
        num_queries = keras.ops.shape(hidden_states)[1]
        sampling_offsets = self.sampling_offsets(hidden_states)
        attention_weights = self.attention_weights(hidden_states)
        attention_weights = keras.ops.softmax(attention_weights, axis=-1)

        if keras.ops.shape(reference_points)[-1] == 2:
            offset_normalizer = keras.ops.cast(
                spatial_shapes, dtype=hidden_states.dtype
            )
            offset_normalizer = keras.ops.flip(offset_normalizer, axis=1)
            offset_normalizer = keras.ops.reshape(
                offset_normalizer, (1, 1, 1, self.n_levels, 1, 2)
            )
            sampling_locations = (
                keras.ops.reshape(
                    reference_points,
                    (batch_size, num_queries, 1, self.n_levels, 1, 2),
                )
                + sampling_offsets / offset_normalizer
            )
        elif keras.ops.shape(reference_points)[-1] == 4:
            num_points_scale_t = keras.ops.cast(
                self.num_points_scale, dtype=hidden_states.dtype
            )
            num_points_scale_t = keras.ops.expand_dims(
                num_points_scale_t, axis=-1
            )
            offset = (
                sampling_offsets
                * num_points_scale_t
                * keras.ops.expand_dims(reference_points[..., 2:], axis=-2)
                * self.offset_scale
            )
            sampling_locations = (
                keras.ops.expand_dims(reference_points[..., :2], axis=-2)
                + offset
            )
        else:
            raise ValueError(
                f"Last dim of reference_points must be 2 or 4, but get "
                f"{keras.ops.shape(reference_points)[-1]} instead."
            )
        return sampling_locations, attention_weights

    def call(
        self,
        hidden_states,
        encoder_hidden_states,
        reference_points,
        spatial_shapes,
    ):
        batch_size = keras.ops.shape(hidden_states)[0]
        num_queries = keras.ops.shape(hidden_states)[1]
        sequence_length = keras.ops.shape(encoder_hidden_states)[1]
        value = keras.ops.reshape(
            encoder_hidden_states,
            (
                batch_size,
                sequence_length,
                self.n_heads,
                self.hidden_dim // self.n_heads,
            ),
        )
        sampling_locations, attention_weights = self.compute_attention(
            hidden_states, reference_points, spatial_shapes
        )

        # NOTE: slice_sizes_values passed down to ms_deformable_attn_core
        # since JAX tracing doesn't support dynamic shapes.
        slice_sizes = [h * w for h, w in self.spatial_shapes]
        output = self.ms_deformable_attn_core(
            value,
            spatial_shapes,
            sampling_locations,
            attention_weights,
            self.num_points,
            slice_sizes,
            self.spatial_shapes,
            self.n_levels,
            num_queries,
            self.decoder_method,
        )
        return output, attention_weights

    def compute_output_spec(
        self,
        hidden_states,
        encoder_hidden_states,
        reference_points,
        spatial_shapes,
    ):
        input_shape = hidden_states.shape
        batch_size = input_shape[0] if len(input_shape) > 0 else None
        num_queries = input_shape[1] if len(input_shape) > 1 else None
        output_shape = (batch_size, num_queries, self.hidden_dim)
        output_spec = keras.KerasTensor(output_shape, dtype=self.compute_dtype)
        attention_weights_shape = (
            batch_size,
            num_queries,
            self.n_heads,
            sum(self.num_points),
        )
        attention_weights_spec = keras.KerasTensor(
            attention_weights_shape, dtype=self.compute_dtype
        )
        return output_spec, attention_weights_spec

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "decoder_attention_heads": self.n_heads,
                "num_feature_levels": self.n_levels,
                "decoder_offset_scale": self.offset_scale,
                "decoder_method": self.decoder_method,
                "decoder_n_points": self.decoder_n_points,
                "num_queries": self.num_queries,
                "spatial_shapes": self.spatial_shapes,
            }
        )
        return config


class DFineMultiheadAttention(keras.layers.Layer):
    """Multi-head attention layer for D-FINE models.

    This layer implements a standard multi-head attention mechanism. It is used
    in two key places within the D-FINE architecture:
    1.  In `DFineEncoderLayer` as the self-attention mechanism to process the
        sequence of image features from the `HGNetV2Backbone` class.
    2.  In `DFineDecoderLayer` as the self-attention mechanism to allow object
        queries to interact with each other.

    It supports position embeddings to incorporate positional information and
    attention masking to prevent attending to certain positions.

    Args:
        embedding_dim: int, Embedding dimension size.
        num_heads: int, Number of attention heads.
        dropout: float, optional, Dropout probability for attention weights.
            Defaults to `0.0`.
        bias: bool, optional, Whether to include bias in projection layers.
            Defaults to `True`.
        kernel_initializer: str or initializer, optional, Initializer for
            kernel weights. Defaults to `"glorot_uniform"`.
        bias_initializer: str or initializer, optional, Initializer for
            bias weights. Defaults to `"zeros"`.
        **kwargs: Additional keyword arguments passed to the parent class.
    """

    def __init__(
        self,
        embedding_dim,
        num_heads,
        dropout=0.0,
        bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.dropout_rate = dropout
        self.head_dim = embedding_dim // num_heads
        if self.head_dim * self.num_heads != self.embedding_dim:
            raise ValueError(
                f"embedding_dim must be divisible by num_heads (got "
                f"`embedding_dim`: {self.embedding_dim} and `num_heads`: "
                f"{self.num_heads})."
            )
        self.scaling = self.head_dim**-0.5
        self.bias = bias
        self.kernel_initializer = keras.initializers.get(kernel_initializer)
        self.bias_initializer = keras.initializers.get(bias_initializer)
        self.dropout = keras.layers.Dropout(
            self.dropout_rate, dtype=self.dtype_policy
        )

    def build(self, input_shape):
        embedding_dim = self.embedding_dim
        proj_equation = "abc,cde->abde"
        proj_bias_axes = "de"
        proj_output_shape = (None, self.num_heads, self.head_dim)
        proj_input_shape = (None, None, embedding_dim)
        self.q_proj = keras.layers.EinsumDense(
            proj_equation,
            output_shape=proj_output_shape,
            bias_axes=proj_bias_axes if self.bias else None,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer if self.bias else None,
            dtype=self.dtype_policy,
            name="q_proj",
        )
        self.q_proj.build(proj_input_shape)
        self.k_proj = keras.layers.EinsumDense(
            proj_equation,
            output_shape=proj_output_shape,
            bias_axes=proj_bias_axes if self.bias else None,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer if self.bias else None,
            dtype=self.dtype_policy,
            name="k_proj",
        )
        self.k_proj.build(proj_input_shape)
        self.v_proj = keras.layers.EinsumDense(
            proj_equation,
            output_shape=proj_output_shape,
            bias_axes=proj_bias_axes if self.bias else None,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer if self.bias else None,
            dtype=self.dtype_policy,
            name="v_proj",
        )
        self.v_proj.build(proj_input_shape)
        out_proj_input_shape = (None, None, self.num_heads * self.head_dim)
        out_proj_output_shape = (None, self.embedding_dim)
        self.out_proj = keras.layers.EinsumDense(
            "abc,cd->abd",
            output_shape=out_proj_output_shape,
            bias_axes="d" if self.bias else None,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer if self.bias else None,
            dtype=self.dtype_policy,
            name="out_proj",
        )
        self.out_proj.build(out_proj_input_shape)
        super().build(input_shape)

    def call(
        self,
        hidden_states,
        position_embeddings=None,
        attention_mask=None,
        output_attentions=False,
        training=None,
    ):
        batch_size = keras.ops.shape(hidden_states)[0]
        target_len = keras.ops.shape(hidden_states)[1]

        def with_pos_embed(tensor, position_embeddings_k):
            return (
                tensor
                if position_embeddings_k is None
                else tensor + position_embeddings_k
            )

        hidden_states_with_pos = with_pos_embed(
            hidden_states, position_embeddings
        )
        query_states = self.q_proj(hidden_states_with_pos)
        key_states = self.k_proj(hidden_states_with_pos)
        value_states = self.v_proj(hidden_states)
        attn_weights = keras.ops.einsum(
            "bthd,bshd->bhts", query_states * self.scaling, key_states
        )
        if attention_mask is not None:
            if keras.ops.ndim(attention_mask) == 2:
                attention_mask = keras.ops.expand_dims(attention_mask, axis=0)
            attention_mask = keras.ops.expand_dims(attention_mask, axis=1)
            attn_weights = attn_weights + attention_mask
        attn_weights = keras.ops.softmax(attn_weights, axis=-1)
        attn_weights_for_output = attn_weights if output_attentions else None
        attn_probs = self.dropout(attn_weights, training=training)
        attn_output = keras.ops.einsum(
            "bhts,bshd->bthd", attn_probs, value_states
        )
        attn_output = keras.ops.reshape(
            attn_output, (batch_size, target_len, self.embedding_dim)
        )
        attn_output = self.out_proj(attn_output)
        if output_attentions:
            return attn_output, attn_weights_for_output
        else:
            return attn_output

    def compute_output_spec(
        self,
        hidden_states,
        position_embeddings=None,
        attention_mask=None,
        output_attentions=False,
        training=None,
    ):
        input_shape = hidden_states.shape
        batch_size = input_shape[0] if len(input_shape) > 0 else None
        target_len = input_shape[1] if len(input_shape) > 1 else None
        source_len = target_len
        attn_output_shape = (batch_size, target_len, self.embedding_dim)
        attn_output_spec = keras.KerasTensor(
            attn_output_shape, dtype=self.compute_dtype
        )
        if output_attentions:
            attn_weights_shape = (
                batch_size,
                self.num_heads,
                target_len,
                source_len,
            )
            attn_weights_spec = keras.KerasTensor(
                attn_weights_shape, dtype=self.compute_dtype
            )
            return attn_output_spec, attn_weights_spec
        return attn_output_spec

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "embedding_dim": self.embedding_dim,
                "num_heads": self.num_heads,
                "dropout": self.dropout_rate,
                "bias": self.bias,
                "kernel_initializer": keras.initializers.serialize(
                    self.kernel_initializer
                ),
                "bias_initializer": keras.initializers.serialize(
                    self.bias_initializer
                ),
            }
        )
        return config



from keras import tree

from keras_hub.src.utils.keras_utils import print_msg

try:
    import openvino as ov
    import openvino.opset14 as ov_opset
    from openvino import Core
except ImportError:
    ov = None
    ov_opset = None
    Core = None


_core = None


def get_core():
    """Get or create OpenVINO Core instance.

    Returns:
        openvino.Core: OpenVINO Core instance,
        or None if OpenVINO not available.
    """
    global _core
    if _core is None and Core is not None:
        _core = Core()
    return _core


def get_device():
    """Detect and return the best available OpenVINO device.

    Returns:
        str: "GPU" if available, otherwise "CPU".
    """
    core = get_core()
    if core is None:
        return "CPU"
    return "GPU" if "GPU" in core.available_devices else "CPU"


def compile_model(struct_params, struct_outputs, device, model_dtype):
    """Compile OpenVINO model with dynamic shapes and precision hints.

    Args:
        struct_params: Model parameters structure.
        struct_outputs: Model outputs structure.
        device: Target device ("GPU" or "CPU").
        model_dtype: Model precision ("f16" or "f32").

    Returns:
        Compiled OpenVINO model ready for inference.
    """
    flat_params = tree.flatten(struct_params)
    flat_outputs = tree.flatten(struct_outputs)
    parameters = [p.output.get_node() for p in flat_params]
    results = [ov_opset.result(r.output) for r in flat_outputs]
    ov_model = ov.Model(results=results, parameters=parameters)
    for ov_input in ov_model.inputs:
        rank = ov_input.get_partial_shape().rank.get_length()
        ov_input.get_node().set_partial_shape(ov.PartialShape([-1] * rank))
    ov_model.validate_nodes_and_infer_types()
    config = {"INFERENCE_PRECISION_HINT": model_dtype}
    core = get_core()
    if core is None:
        raise RuntimeError("OpenVINO not available")
    return core.compile_model(ov_model, device, config)


def get_outputs(inputs, struct_outputs, compiled_ov_model, unpack_singleton):
    """Execute compiled OpenVINO model and return structured outputs.

    Args:
        inputs: Input tensors for inference.
        struct_outputs: Expected output structure.
        compiled_ov_model: Compiled OpenVINO model.
        unpack_singleton: Function to unpack singleton outputs.

    Returns:
        Structured model outputs matching expected format.
    """
    flatten_inputs = tree.flatten(inputs)
    raw = compiled_ov_model(flatten_inputs).to_tuple()
    packed = tree.pack_sequence_as(struct_outputs, raw)
    return unpack_singleton(packed)


def ov_infer(model, inputs, stop_token_ids, fn):
    """High-level OpenVINO inference with model reuse and compilation.

    This function manages OpenVINO model compilation and caching. It reuses
    existing compiled models when possible, or compiles new ones as needed.
    Handles device detection and automatic precision selection.

    Args:
        model: Keras model with OpenVINO backend support.
        inputs: Input tensors for inference.
        stop_token_ids: Token IDs that should stop generation.
        fn: Function to execute with the parameterized inputs.

    Returns:
        Model outputs from OpenVINO inference.
    """
    device = get_device()

    # Try to use existing compiled model for the same device
    if (
        getattr(model, "ov_compiled_model", None) is not None
        and getattr(model, "ov_device", None) is not None
        and device == model.ov_device
    ):
        try:
            return get_outputs(
                inputs,
                model.struct_outputs,
                model.ov_compiled_model,
                model._unpack_singleton,
            )
        except RuntimeError as e:
            print_msg(
                "WARNING: OpenVINO inference \033[1mFAILED\033[0m, "
                "recompiling model and trying again.\n" + str(e)
            )
            model.ov_compiled_model = None
            model.struct_outputs = None

    # Compile a new model
    struct_params = model._parameterize_data(inputs)
    model.struct_outputs = fn(struct_params, stop_token_ids)
    model.ov_device = device
    model_dtype = "f16" if model.dtype in ("float16", "bfloat16") else "f32"
    model.ov_compiled_model = compile_model(
        struct_params, model.struct_outputs, device, model_dtype
    )
    return get_outputs(
        inputs,
        model.struct_outputs,
        model.ov_compiled_model,
        model._unpack_singleton,
    )



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.src.models.albert.albert_backbone import (
    AlbertBackbone as AlbertBackbone,
)
from keras_hub.src.models.albert.albert_masked_lm import (
    AlbertMaskedLM as AlbertMaskedLM,
)
from keras_hub.src.models.albert.albert_masked_lm_preprocessor import (
    AlbertMaskedLMPreprocessor as AlbertMaskedLMPreprocessor,
)
from keras_hub.src.models.albert.albert_text_classifier import (
    AlbertTextClassifier as AlbertClassifier,
)
from keras_hub.src.models.albert.albert_text_classifier import (
    AlbertTextClassifier as AlbertTextClassifier,
)
from keras_hub.src.models.albert.albert_text_classifier_preprocessor import (
    AlbertTextClassifierPreprocessor as AlbertPreprocessor,
)
from keras_hub.src.models.albert.albert_text_classifier_preprocessor import (
    AlbertTextClassifierPreprocessor as AlbertTextClassifierPreprocessor,
)
from keras_hub.src.models.albert.albert_tokenizer import (
    AlbertTokenizer as AlbertTokenizer,
)
from keras_hub.src.models.backbone import Backbone as Backbone
from keras_hub.src.models.bart.bart_backbone import BartBackbone as BartBackbone
from keras_hub.src.models.bart.bart_seq_2_seq_lm import (
    BartSeq2SeqLM as BartSeq2SeqLM,
)
from keras_hub.src.models.bart.bart_seq_2_seq_lm_preprocessor import (
    BartSeq2SeqLMPreprocessor as BartSeq2SeqLMPreprocessor,
)
from keras_hub.src.models.bart.bart_tokenizer import (
    BartTokenizer as BartTokenizer,
)
from keras_hub.src.models.basnet.basnet import (
    BASNetImageSegmenter as BASNetImageSegmenter,
)
from keras_hub.src.models.basnet.basnet_backbone import (
    BASNetBackbone as BASNetBackbone,
)
from keras_hub.src.models.basnet.basnet_preprocessor import (
    BASNetPreprocessor as BASNetPreprocessor,
)
from keras_hub.src.models.bert.bert_backbone import BertBackbone as BertBackbone
from keras_hub.src.models.bert.bert_masked_lm import (
    BertMaskedLM as BertMaskedLM,
)
from keras_hub.src.models.bert.bert_masked_lm_preprocessor import (
    BertMaskedLMPreprocessor as BertMaskedLMPreprocessor,
)
from keras_hub.src.models.bert.bert_text_classifier import (
    BertTextClassifier as BertClassifier,
)
from keras_hub.src.models.bert.bert_text_classifier import (
    BertTextClassifier as BertTextClassifier,
)
from keras_hub.src.models.bert.bert_text_classifier_preprocessor import (
    BertTextClassifierPreprocessor as BertPreprocessor,
)
from keras_hub.src.models.bert.bert_text_classifier_preprocessor import (
    BertTextClassifierPreprocessor as BertTextClassifierPreprocessor,
)
from keras_hub.src.models.bert.bert_tokenizer import (
    BertTokenizer as BertTokenizer,
)
from keras_hub.src.models.bloom.bloom_backbone import (
    BloomBackbone as BloomBackbone,
)
from keras_hub.src.models.bloom.bloom_causal_lm import (
    BloomCausalLM as BloomCausalLM,
)
from keras_hub.src.models.bloom.bloom_causal_lm_preprocessor import (
    BloomCausalLMPreprocessor as BloomCausalLMPreprocessor,
)
from keras_hub.src.models.bloom.bloom_tokenizer import (
    BloomTokenizer as BloomTokenizer,
)
from keras_hub.src.models.causal_lm import CausalLM as CausalLM
from keras_hub.src.models.causal_lm_preprocessor import (
    CausalLMPreprocessor as CausalLMPreprocessor,
)
from keras_hub.src.models.clip.clip_backbone import CLIPBackbone as CLIPBackbone
from keras_hub.src.models.clip.clip_preprocessor import (
    CLIPPreprocessor as CLIPPreprocessor,
)
from keras_hub.src.models.clip.clip_text_encoder import (
    CLIPTextEncoder as CLIPTextEncoder,
)
from keras_hub.src.models.clip.clip_tokenizer import (
    CLIPTokenizer as CLIPTokenizer,
)
from keras_hub.src.models.clip.clip_vision_encoder import (
    CLIPVisionEncoder as CLIPVisionEncoder,
)
from keras_hub.src.models.cspnet.cspnet_backbone import (
    CSPNetBackbone as CSPNetBackbone,
)
from keras_hub.src.models.cspnet.cspnet_image_classifier import (
    CSPNetImageClassifier as CSPNetImageClassifier,
)
from keras_hub.src.models.cspnet.cspnet_image_classifier_preprocessor import (
    CSPNetImageClassifierPreprocessor as CSPNetImageClassifierPreprocessor,
)
from keras_hub.src.models.d_fine.d_fine_backbone import (
    DFineBackbone as DFineBackbone,
)
from keras_hub.src.models.d_fine.d_fine_object_detector import (
    DFineObjectDetector as DFineObjectDetector,
)
from keras_hub.src.models.d_fine.d_fine_object_detector_preprocessor import (
    DFineObjectDetectorPreprocessor as DFineObjectDetectorPreprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_backbone import (
    DebertaV3Backbone as DebertaV3Backbone,
)
from keras_hub.src.models.deberta_v3.deberta_v3_masked_lm import (
    DebertaV3MaskedLM as DebertaV3MaskedLM,
)
from keras_hub.src.models.deberta_v3.deberta_v3_masked_lm_preprocessor import (
    DebertaV3MaskedLMPreprocessor as DebertaV3MaskedLMPreprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier import (
    DebertaV3TextClassifier as DebertaV3Classifier,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier import (
    DebertaV3TextClassifier as DebertaV3TextClassifier,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier_preprocessor import (
    DebertaV3TextClassifierPreprocessor as DebertaV3Preprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier_preprocessor import (
    DebertaV3TextClassifierPreprocessor as DebertaV3TextClassifierPreprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_tokenizer import (
    DebertaV3Tokenizer as DebertaV3Tokenizer,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_backbone import (
    DeepLabV3Backbone as DeepLabV3Backbone,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_image_segmeter_preprocessor import (
    DeepLabV3ImageSegmenterPreprocessor as DeepLabV3ImageSegmenterPreprocessor,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_segmenter import (
    DeepLabV3ImageSegmenter as DeepLabV3ImageSegmenter,
)
from keras_hub.src.models.deit.deit_backbone import DeiTBackbone as DeiTBackbone
from keras_hub.src.models.deit.deit_image_classifier import (
    DeiTImageClassifier as DeiTImageClassifier,
)
from keras_hub.src.models.deit.deit_image_classifier_preprocessor import (
    DeiTImageClassifierPreprocessor as DeiTImageClassifierPreprocessor,
)
from keras_hub.src.models.densenet.densenet_backbone import (
    DenseNetBackbone as DenseNetBackbone,
)
from keras_hub.src.models.densenet.densenet_image_classifier import (
    DenseNetImageClassifier as DenseNetImageClassifier,
)
from keras_hub.src.models.densenet.densenet_image_classifier_preprocessor import (
    DenseNetImageClassifierPreprocessor as DenseNetImageClassifierPreprocessor,
)
from keras_hub.src.models.depth_anything.depth_anything_backbone import (
    DepthAnythingBackbone as DepthAnythingBackbone,
)
from keras_hub.src.models.depth_anything.depth_anything_depth_estimator import (
    DepthAnythingDepthEstimator as DepthAnythingDepthEstimator,
)
from keras_hub.src.models.depth_anything.depth_anything_depth_estimator_preprocessor import (
    DepthAnythingDepthEstimatorPreprocessor as DepthAnythingDepthEstimatorPreprocessor,
)
from keras_hub.src.models.depth_estimator import (
    DepthEstimator as DepthEstimator,
)
from keras_hub.src.models.depth_estimator_preprocessor import (
    DepthEstimatorPreprocessor as DepthEstimatorPreprocessor,
)
from keras_hub.src.models.dinov2.dinov2_backbone import (
    DINOV2Backbone as DINOV2Backbone,
)
from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    DistilBertBackbone as DistilBertBackbone,
)
from keras_hub.src.models.distil_bert.distil_bert_masked_lm import (
    DistilBertMaskedLM as DistilBertMaskedLM,
)
from keras_hub.src.models.distil_bert.distil_bert_masked_lm_preprocessor import (
    DistilBertMaskedLMPreprocessor as DistilBertMaskedLMPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier import (
    DistilBertTextClassifier as DistilBertClassifier,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier import (
    DistilBertTextClassifier as DistilBertTextClassifier,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier_preprocessor import (
    DistilBertTextClassifierPreprocessor as DistilBertPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier_preprocessor import (
    DistilBertTextClassifierPreprocessor as DistilBertTextClassifierPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_tokenizer import (
    DistilBertTokenizer as DistilBertTokenizer,
)
from keras_hub.src.models.efficientnet.efficientnet_backbone import (
    EfficientNetBackbone as EfficientNetBackbone,
)
from keras_hub.src.models.efficientnet.efficientnet_image_classifier import (
    EfficientNetImageClassifier as EfficientNetImageClassifier,
)
from keras_hub.src.models.efficientnet.efficientnet_image_classifier_preprocessor import (
    EfficientNetImageClassifierPreprocessor as EfficientNetImageClassifierPreprocessor,
)
from keras_hub.src.models.electra.electra_backbone import (
    ElectraBackbone as ElectraBackbone,
)
from keras_hub.src.models.electra.electra_tokenizer import (
    ElectraTokenizer as ElectraTokenizer,
)
from keras_hub.src.models.esm.esm_backbone import ESMBackbone as ESM2Backbone
from keras_hub.src.models.esm.esm_backbone import ESMBackbone as ESMBackbone
from keras_hub.src.models.esm.esm_classifier import (
    ESMProteinClassifier as ESMProteinClassifier,
)
from keras_hub.src.models.esm.esm_classifier_preprocessor import (
    ESMProteinClassifierPreprocessor as ESMProteinClassifierPreprocessor,
)
from keras_hub.src.models.esm.esm_masked_plm import (
    ESMMaskedPLM as ESM2MaskedPLM,
)
from keras_hub.src.models.esm.esm_masked_plm import ESMMaskedPLM as ESMMaskedPLM
from keras_hub.src.models.esm.esm_masked_plm_preprocessor import (
    ESMMaskedPLMPreprocessor as ESMMaskedPLMPreprocessor,
)
from keras_hub.src.models.esm.esm_tokenizer import ESMTokenizer as ESMTokenizer
from keras_hub.src.models.f_net.f_net_backbone import (
    FNetBackbone as FNetBackbone,
)
from keras_hub.src.models.f_net.f_net_masked_lm import (
    FNetMaskedLM as FNetMaskedLM,
)
from keras_hub.src.models.f_net.f_net_masked_lm_preprocessor import (
    FNetMaskedLMPreprocessor as FNetMaskedLMPreprocessor,
)
from keras_hub.src.models.f_net.f_net_text_classifier import (
    FNetTextClassifier as FNetClassifier,
)
from keras_hub.src.models.f_net.f_net_text_classifier import (
    FNetTextClassifier as FNetTextClassifier,
)
from keras_hub.src.models.f_net.f_net_text_classifier_preprocessor import (
    FNetTextClassifierPreprocessor as FNetPreprocessor,
)
from keras_hub.src.models.f_net.f_net_text_classifier_preprocessor import (
    FNetTextClassifierPreprocessor as FNetTextClassifierPreprocessor,
)
from keras_hub.src.models.f_net.f_net_tokenizer import (
    FNetTokenizer as FNetTokenizer,
)
from keras_hub.src.models.falcon.falcon_backbone import (
    FalconBackbone as FalconBackbone,
)
from keras_hub.src.models.falcon.falcon_causal_lm import (
    FalconCausalLM as FalconCausalLM,
)
from keras_hub.src.models.falcon.falcon_causal_lm_preprocessor import (
    FalconCausalLMPreprocessor as FalconCausalLMPreprocessor,
)
from keras_hub.src.models.falcon.falcon_tokenizer import (
    FalconTokenizer as FalconTokenizer,
)
from keras_hub.src.models.feature_pyramid_backbone import (
    FeaturePyramidBackbone as FeaturePyramidBackbone,
)
from keras_hub.src.models.flux.flux_model import FluxBackbone as FluxBackbone
from keras_hub.src.models.flux.flux_text_to_image import (
    FluxTextToImage as FluxTextToImage,
)
from keras_hub.src.models.flux.flux_text_to_image_preprocessor import (
    FluxTextToImagePreprocessor as FluxTextToImagePreprocessor,
)
from keras_hub.src.models.gemma.gemma_backbone import (
    GemmaBackbone as GemmaBackbone,
)
from keras_hub.src.models.gemma.gemma_causal_lm import (
    GemmaCausalLM as GemmaCausalLM,
)
from keras_hub.src.models.gemma.gemma_causal_lm_preprocessor import (
    GemmaCausalLMPreprocessor as GemmaCausalLMPreprocessor,
)
from keras_hub.src.models.gemma.gemma_tokenizer import (
    GemmaTokenizer as GemmaTokenizer,
)
from keras_hub.src.models.gemma3.gemma3_backbone import (
    Gemma3Backbone as Gemma3Backbone,
)
from keras_hub.src.models.gemma3.gemma3_causal_lm import (
    Gemma3CausalLM as Gemma3CausalLM,
)
from keras_hub.src.models.gemma3.gemma3_causal_lm_preprocessor import (
    Gemma3CausalLMPreprocessor as Gemma3CausalLMPreprocessor,
)
from keras_hub.src.models.gemma3.gemma3_tokenizer import (
    Gemma3Tokenizer as Gemma3Tokenizer,
)
from keras_hub.src.models.gemma3.gemma3_vision_encoder import (
    Gemma3VisionEncoder as Gemma3VisionEncoder,
)
from keras_hub.src.models.gpt2.gpt2_backbone import GPT2Backbone as GPT2Backbone
from keras_hub.src.models.gpt2.gpt2_causal_lm import (
    GPT2CausalLM as GPT2CausalLM,
)
from keras_hub.src.models.gpt2.gpt2_causal_lm_preprocessor import (
    GPT2CausalLMPreprocessor as GPT2CausalLMPreprocessor,
)
from keras_hub.src.models.gpt2.gpt2_preprocessor import (
    GPT2Preprocessor as GPT2Preprocessor,
)
from keras_hub.src.models.gpt2.gpt2_tokenizer import (
    GPT2Tokenizer as GPT2Tokenizer,
)
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_backbone import (
    GPTNeoXBackbone as GPTNeoXBackbone,
)
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_causal_lm import (
    GPTNeoXCausalLM as GPTNeoXCausalLM,
)
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_causal_lm_preprocessor import (
    GPTNeoXCausalLMPreprocessor as GPTNeoXCausalLMPreprocessor,
)
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_tokenizer import (
    GPTNeoXTokenizer as GPTNeoXTokenizer,
)
from keras_hub.src.models.hgnetv2.hgnetv2_backbone import (
    HGNetV2Backbone as HGNetV2Backbone,
)
from keras_hub.src.models.hgnetv2.hgnetv2_image_classifier import (
    HGNetV2ImageClassifier as HGNetV2ImageClassifier,
)
from keras_hub.src.models.hgnetv2.hgnetv2_image_classifier_preprocessor import (
    HGNetV2ImageClassifierPreprocessor as HGNetV2ImageClassifierPreprocessor,
)
from keras_hub.src.models.image_classifier import (
    ImageClassifier as ImageClassifier,
)
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor as ImageClassifierPreprocessor,
)
from keras_hub.src.models.image_segmenter import (
    ImageSegmenter as ImageSegmenter,
)
from keras_hub.src.models.image_segmenter_preprocessor import (
    ImageSegmenterPreprocessor as ImageSegmenterPreprocessor,
)
from keras_hub.src.models.image_to_image import ImageToImage as ImageToImage
from keras_hub.src.models.inpaint import Inpaint as Inpaint
from keras_hub.src.models.llama.llama_backbone import (
    LlamaBackbone as LlamaBackbone,
)
from keras_hub.src.models.llama.llama_causal_lm import (
    LlamaCausalLM as LlamaCausalLM,
)
from keras_hub.src.models.llama.llama_causal_lm_preprocessor import (
    LlamaCausalLMPreprocessor as LlamaCausalLMPreprocessor,
)
from keras_hub.src.models.llama.llama_tokenizer import (
    LlamaTokenizer as LlamaTokenizer,
)
from keras_hub.src.models.llama3.llama3_backbone import (
    Llama3Backbone as Llama3Backbone,
)
from keras_hub.src.models.llama3.llama3_causal_lm import (
    Llama3CausalLM as Llama3CausalLM,
)
from keras_hub.src.models.llama3.llama3_causal_lm_preprocessor import (
    Llama3CausalLMPreprocessor as Llama3CausalLMPreprocessor,
)
from keras_hub.src.models.llama3.llama3_tokenizer import (
    Llama3Tokenizer as Llama3Tokenizer,
)
from keras_hub.src.models.masked_lm import MaskedLM as MaskedLM
from keras_hub.src.models.masked_lm_preprocessor import (
    MaskedLMPreprocessor as MaskedLMPreprocessor,
)
from keras_hub.src.models.mistral.mistral_backbone import (
    MistralBackbone as MistralBackbone,
)
from keras_hub.src.models.mistral.mistral_causal_lm import (
    MistralCausalLM as MistralCausalLM,
)
from keras_hub.src.models.mistral.mistral_causal_lm_preprocessor import (
    MistralCausalLMPreprocessor as MistralCausalLMPreprocessor,
)
from keras_hub.src.models.mistral.mistral_tokenizer import (
    MistralTokenizer as MistralTokenizer,
)
from keras_hub.src.models.mit.mit_backbone import MiTBackbone as MiTBackbone
from keras_hub.src.models.mit.mit_image_classifier import (
    MiTImageClassifier as MiTImageClassifier,
)
from keras_hub.src.models.mit.mit_image_classifier_preprocessor import (
    MiTImageClassifierPreprocessor as MiTImageClassifierPreprocessor,
)
from keras_hub.src.models.mixtral.mixtral_backbone import (
    MixtralBackbone as MixtralBackbone,
)
from keras_hub.src.models.mixtral.mixtral_causal_lm import (
    MixtralCausalLM as MixtralCausalLM,
)
from keras_hub.src.models.mixtral.mixtral_causal_lm_preprocessor import (
    MixtralCausalLMPreprocessor as MixtralCausalLMPreprocessor,
)
from keras_hub.src.models.mixtral.mixtral_tokenizer import (
    MixtralTokenizer as MixtralTokenizer,
)
from keras_hub.src.models.mobilenet.mobilenet_backbone import (
    MobileNetBackbone as MobileNetBackbone,
)
from keras_hub.src.models.mobilenet.mobilenet_image_classifier import (
    MobileNetImageClassifier as MobileNetImageClassifier,
)
from keras_hub.src.models.mobilenet.mobilenet_image_classifier_preprocessor import (
    MobileNetImageClassifierPreprocessor as MobileNetImageClassifierPreprocessor,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_backbone import (
    MobileNetV5Backbone as MobileNetV5Backbone,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_image_classifier import (
    MobileNetV5ImageClassifier as MobileNetV5ImageClassifier,
)
from keras_hub.src.models.mobilenetv5.mobilenetv5_image_classifier_preprocessor import (
    MobileNetV5ImageClassifierPreprocessor as MobileNetV5ImageClassifierPreprocessor,
)
from keras_hub.src.models.moonshine.moonshine_audio_to_text import (
    MoonshineAudioToText as MoonshineAudioToText,
)
from keras_hub.src.models.moonshine.moonshine_audio_to_text_preprocessor import (
    MoonshineAudioToTextPreprocessor as MoonshineAudioToTextPreprocessor,
)
from keras_hub.src.models.moonshine.moonshine_backbone import (
    MoonshineBackbone as MoonshineBackbone,
)
from keras_hub.src.models.moonshine.moonshine_tokenizer import (
    MoonshineTokenizer as MoonshineTokenizer,
)
from keras_hub.src.models.object_detector import (
    ObjectDetector as ImageObjectDetector,
)
from keras_hub.src.models.object_detector import (
    ObjectDetector as ObjectDetector,
)
from keras_hub.src.models.object_detector_preprocessor import (
    ObjectDetectorPreprocessor as ImageObjectDetectorPreprocessor,
)
from keras_hub.src.models.object_detector_preprocessor import (
    ObjectDetectorPreprocessor as ObjectDetectorPreprocessor,
)
from keras_hub.src.models.opt.opt_backbone import OPTBackbone as OPTBackbone
from keras_hub.src.models.opt.opt_causal_lm import OPTCausalLM as OPTCausalLM
from keras_hub.src.models.opt.opt_causal_lm_preprocessor import (
    OPTCausalLMPreprocessor as OPTCausalLMPreprocessor,
)
from keras_hub.src.models.opt.opt_tokenizer import OPTTokenizer as OPTTokenizer
from keras_hub.src.models.pali_gemma.pali_gemma_backbone import (
    PaliGemmaBackbone as PaliGemmaBackbone,
)
from keras_hub.src.models.pali_gemma.pali_gemma_causal_lm import (
    PaliGemmaCausalLM as PaliGemmaCausalLM,
)
from keras_hub.src.models.pali_gemma.pali_gemma_causal_lm_preprocessor import (
    PaliGemmaCausalLMPreprocessor as PaliGemmaCausalLMPreprocessor,
)
from keras_hub.src.models.pali_gemma.pali_gemma_tokenizer import (
    PaliGemmaTokenizer as PaliGemmaTokenizer,
)
from keras_hub.src.models.parseq.parseq_backbone import (
    PARSeqBackbone as PARSeqBackbone,
)
from keras_hub.src.models.parseq.parseq_causal_lm import (
    PARSeqCausalLM as PARSeqCausalLM,
)
from keras_hub.src.models.parseq.parseq_causal_lm_preprocessor import (
    PARSeqCausalLMPreprocessor as PARSeqCausalLMPreprocessor,
)
from keras_hub.src.models.parseq.parseq_tokenizer import (
    PARSeqTokenizer as PARSeqTokenizer,
)
from keras_hub.src.models.phi3.phi3_backbone import Phi3Backbone as Phi3Backbone
from keras_hub.src.models.phi3.phi3_causal_lm import (
    Phi3CausalLM as Phi3CausalLM,
)
from keras_hub.src.models.phi3.phi3_causal_lm_preprocessor import (
    Phi3CausalLMPreprocessor as Phi3CausalLMPreprocessor,
)
from keras_hub.src.models.phi3.phi3_tokenizer import (
    Phi3Tokenizer as Phi3Tokenizer,
)
from keras_hub.src.models.preprocessor import Preprocessor as Preprocessor
from keras_hub.src.models.qwen.qwen_backbone import (
    QwenBackbone as Qwen2Backbone,
)
from keras_hub.src.models.qwen.qwen_backbone import QwenBackbone as QwenBackbone
from keras_hub.src.models.qwen.qwen_causal_lm import (
    QwenCausalLM as Qwen2CausalLM,
)
from keras_hub.src.models.qwen.qwen_causal_lm import (
    QwenCausalLM as QwenCausalLM,
)
from keras_hub.src.models.qwen.qwen_causal_lm_preprocessor import (
    QwenCausalLMPreprocessor as Qwen2CausalLMPreprocessor,
)
from keras_hub.src.models.qwen.qwen_causal_lm_preprocessor import (
    QwenCausalLMPreprocessor as QwenCausalLMPreprocessor,
)
from keras_hub.src.models.qwen.qwen_tokenizer import (
    QwenTokenizer as Qwen2Tokenizer,
)
from keras_hub.src.models.qwen.qwen_tokenizer import (
    QwenTokenizer as QwenTokenizer,
)
from keras_hub.src.models.qwen3.qwen3_backbone import (
    Qwen3Backbone as Qwen3Backbone,
)
from keras_hub.src.models.qwen3.qwen3_causal_lm import (
    Qwen3CausalLM as Qwen3CausalLM,
)
from keras_hub.src.models.qwen3.qwen3_causal_lm_preprocessor import (
    Qwen3CausalLMPreprocessor as Qwen3CausalLMPreprocessor,
)
from keras_hub.src.models.qwen3.qwen3_tokenizer import (
    Qwen3Tokenizer as Qwen3Tokenizer,
)
from keras_hub.src.models.qwen3_moe.qwen3_moe_backbone import (
    Qwen3MoeBackbone as Qwen3MoeBackbone,
)
from keras_hub.src.models.qwen3_moe.qwen3_moe_causal_lm import (
    Qwen3MoeCausalLM as Qwen3MoeCausalLM,
)
from keras_hub.src.models.qwen3_moe.qwen3_moe_causal_lm_preprocessor import (
    Qwen3MoeCausalLMPreprocessor as Qwen3MoeCausalLMPreprocessor,
)
from keras_hub.src.models.qwen_moe.qwen_moe_backbone import (
    QwenMoeBackbone as QwenMoeBackbone,
)
from keras_hub.src.models.qwen_moe.qwen_moe_causal_lm import (
    QwenMoeCausalLM as QwenMoeCausalLM,
)
from keras_hub.src.models.qwen_moe.qwen_moe_causal_lm_preprocessor import (
    QwenMoeCausalLMPreprocessor as QwenMoeCausalLMPreprocessor,
)
from keras_hub.src.models.resnet.resnet_backbone import (
    ResNetBackbone as ResNetBackbone,
)
from keras_hub.src.models.resnet.resnet_image_classifier import (
    ResNetImageClassifier as ResNetImageClassifier,
)
from keras_hub.src.models.resnet.resnet_image_classifier_preprocessor import (
    ResNetImageClassifierPreprocessor as ResNetImageClassifierPreprocessor,
)
from keras_hub.src.models.retinanet.retinanet_backbone import (
    RetinaNetBackbone as RetinaNetBackbone,
)
from keras_hub.src.models.retinanet.retinanet_object_detector import (
    RetinaNetObjectDetector as RetinaNetObjectDetector,
)
from keras_hub.src.models.retinanet.retinanet_object_detector_preprocessor import (
    RetinaNetObjectDetectorPreprocessor as RetinaNetObjectDetectorPreprocessor,
)
from keras_hub.src.models.roberta.roberta_backbone import (
    RobertaBackbone as RobertaBackbone,
)
from keras_hub.src.models.roberta.roberta_masked_lm import (
    RobertaMaskedLM as RobertaMaskedLM,
)
from keras_hub.src.models.roberta.roberta_masked_lm_preprocessor import (
    RobertaMaskedLMPreprocessor as RobertaMaskedLMPreprocessor,
)
from keras_hub.src.models.roberta.roberta_text_classifier import (
    RobertaTextClassifier as RobertaClassifier,
)
from keras_hub.src.models.roberta.roberta_text_classifier import (
    RobertaTextClassifier as RobertaTextClassifier,
)
from keras_hub.src.models.roberta.roberta_text_classifier_preprocessor import (
    RobertaTextClassifierPreprocessor as RobertaPreprocessor,
)
from keras_hub.src.models.roberta.roberta_text_classifier_preprocessor import (
    RobertaTextClassifierPreprocessor as RobertaTextClassifierPreprocessor,
)
from keras_hub.src.models.roberta.roberta_tokenizer import (
    RobertaTokenizer as RobertaTokenizer,
)
from keras_hub.src.models.roformer_v2.roformer_v2_backbone import (
    RoformerV2Backbone as RoformerV2Backbone,
)
from keras_hub.src.models.roformer_v2.roformer_v2_masked_lm import (
    RoformerV2MaskedLM as RoformerV2MaskedLM,
)
from keras_hub.src.models.roformer_v2.roformer_v2_masked_lm_preprocessor import (
    RoformerV2MaskedLMPreprocessor as RoformerV2MaskedLMPreprocessor,
)
from keras_hub.src.models.roformer_v2.roformer_v2_text_classifier import (
    RoformerV2TextClassifier as RoformerV2TextClassifier,
)
from keras_hub.src.models.roformer_v2.roformer_v2_text_classifier_preprocessor import (
    RoformerV2TextClassifierPreprocessor as RoformerV2TextClassifierPreprocessor,
)
from keras_hub.src.models.roformer_v2.roformer_v2_tokenizer import (
    RoformerV2Tokenizer as RoformerV2Tokenizer,
)
from keras_hub.src.models.sam.sam_backbone import SAMBackbone as SAMBackbone
from keras_hub.src.models.sam.sam_image_segmenter import (
    SAMImageSegmenter as SAMImageSegmenter,
)
from keras_hub.src.models.sam.sam_image_segmenter_preprocessor import (
    SAMImageSegmenterPreprocessor as SAMImageSegmenterPreprocessor,
)
from keras_hub.src.models.segformer.segformer_backbone import (
    SegFormerBackbone as SegFormerBackbone,
)
from keras_hub.src.models.segformer.segformer_image_segmenter import (
    SegFormerImageSegmenter as SegFormerImageSegmenter,
)
from keras_hub.src.models.segformer.segformer_image_segmenter_preprocessor import (
    SegFormerImageSegmenterPreprocessor as SegFormerImageSegmenterPreprocessor,
)
from keras_hub.src.models.seq_2_seq_lm import Seq2SeqLM as Seq2SeqLM
from keras_hub.src.models.seq_2_seq_lm_preprocessor import (
    Seq2SeqLMPreprocessor as Seq2SeqLMPreprocessor,
)
from keras_hub.src.models.siglip.siglip_backbone import (
    SigLIPBackbone as SigLIPBackbone,
)
from keras_hub.src.models.siglip.siglip_preprocessor import (
    SigLIPPreprocessor as SigLIPPreprocessor,
)
from keras_hub.src.models.siglip.siglip_text_encoder import (
    SigLIPTextEncoder as SigLIPTextEncoder,
)
from keras_hub.src.models.siglip.siglip_tokenizer import (
    SigLIPTokenizer as SigLIPTokenizer,
)
from keras_hub.src.models.siglip.siglip_vision_encoder import (
    SigLIPVisionEncoder as SigLIPVisionEncoder,
)
from keras_hub.src.models.smollm3.smollm3_backbone import (
    SmolLM3Backbone as SmolLM3Backbone,
)
from keras_hub.src.models.smollm3.smollm3_backbone import (
    SmolLM3Backbone as SmolLMBackbone,
)
from keras_hub.src.models.smollm3.smollm3_causal_lm import (
    SmolLM3CausalLM as SmolLM3CausalLM,
)
from keras_hub.src.models.smollm3.smollm3_causal_lm import (
    SmolLM3CausalLM as SmolLMCausalLM,
)
from keras_hub.src.models.smollm3.smollm3_causal_lm_preprocessor import (
    SmolLM3CausalLMPreprocessor as SmolLM3CausalLMPreprocessor,
)
from keras_hub.src.models.smollm3.smollm3_causal_lm_preprocessor import (
    SmolLM3CausalLMPreprocessor as SmolLMCausalLMPreprocessor,
)
from keras_hub.src.models.smollm3.smollm3_tokenizer import (
    SmolLM3Tokenizer as SmolLM3Tokenizer,
)
from keras_hub.src.models.smollm3.smollm3_tokenizer import (
    SmolLM3Tokenizer as SmolLMTokenizer,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_backbone import (
    StableDiffusion3Backbone as StableDiffusion3Backbone,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_image_to_image import (
    StableDiffusion3ImageToImage as StableDiffusion3ImageToImage,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_inpaint import (
    StableDiffusion3Inpaint as StableDiffusion3Inpaint,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_text_to_image import (
    StableDiffusion3TextToImage as StableDiffusion3TextToImage,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_text_to_image_preprocessor import (
    StableDiffusion3TextToImagePreprocessor as StableDiffusion3TextToImagePreprocessor,
)
from keras_hub.src.models.t5.t5_backbone import T5Backbone as T5Backbone
from keras_hub.src.models.t5.t5_preprocessor import (
    T5Preprocessor as T5Preprocessor,
)
from keras_hub.src.models.t5.t5_tokenizer import T5Tokenizer as T5Tokenizer
from keras_hub.src.models.t5gemma.t5gemma_backbone import (
    T5GemmaBackbone as T5GemmaBackbone,
)
from keras_hub.src.models.t5gemma.t5gemma_seq_2_seq_lm import (
    T5GemmaSeq2SeqLM as T5GemmaSeq2SeqLM,
)
from keras_hub.src.models.t5gemma.t5gemma_seq_2_seq_lm_preprocessor import (
    T5GemmaSeq2SeqLMPreprocessor as T5GemmaSeq2SeqLMPreprocessor,
)
from keras_hub.src.models.t5gemma.t5gemma_tokenizer import (
    T5GemmaTokenizer as T5GemmaTokenizer,
)
from keras_hub.src.models.task import Task as Task
from keras_hub.src.models.text_classifier import TextClassifier as Classifier
from keras_hub.src.models.text_classifier import (
    TextClassifier as TextClassifier,
)
from keras_hub.src.models.text_classifier_preprocessor import (
    TextClassifierPreprocessor as TextClassifierPreprocessor,
)
from keras_hub.src.models.text_to_image import TextToImage as TextToImage
from keras_hub.src.models.text_to_image_preprocessor import (
    TextToImagePreprocessor as TextToImagePreprocessor,
)
from keras_hub.src.models.vgg.vgg_backbone import VGGBackbone as VGGBackbone
from keras_hub.src.models.vgg.vgg_image_classifier import (
    VGGImageClassifier as VGGImageClassifier,
)
from keras_hub.src.models.vgg.vgg_image_classifier_preprocessor import (
    VGGImageClassifierPreprocessor as VGGImageClassifierPreprocessor,
)
from keras_hub.src.models.vit.vit_backbone import ViTBackbone as ViTBackbone
from keras_hub.src.models.vit.vit_image_classifier import (
    ViTImageClassifier as ViTImageClassifier,
)
from keras_hub.src.models.vit.vit_image_classifier_preprocessor import (
    ViTImageClassifierPreprocessor as ViTImageClassifierPreprocessor,
)
from keras_hub.src.models.vit_det.vit_det_backbone import (
    ViTDetBackbone as ViTDetBackbone,
)
from keras_hub.src.models.whisper.whisper_backbone import (
    WhisperBackbone as WhisperBackbone,
)
from keras_hub.src.models.whisper.whisper_tokenizer import (
    WhisperTokenizer as WhisperTokenizer,
)
from keras_hub.src.models.xception.xception_backbone import (
    XceptionBackbone as XceptionBackbone,
)
from keras_hub.src.models.xception.xception_image_classifier import (
    XceptionImageClassifier as XceptionImageClassifier,
)
from keras_hub.src.models.xception.xception_image_classifier_preprocessor import (
    XceptionImageClassifierPreprocessor as XceptionImageClassifierPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_backbone import (
    XLMRobertaBackbone as XLMRobertaBackbone,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_masked_lm import (
    XLMRobertaMaskedLM as XLMRobertaMaskedLM,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_masked_lm_preprocessor import (
    XLMRobertaMaskedLMPreprocessor as XLMRobertaMaskedLMPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier import (
    XLMRobertaTextClassifier as XLMRobertaClassifier,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier import (
    XLMRobertaTextClassifier as XLMRobertaTextClassifier,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier_preprocessor import (
    XLMRobertaTextClassifierPreprocessor as XLMRobertaPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier_preprocessor import (
    XLMRobertaTextClassifierPreprocessor as XLMRobertaTextClassifierPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_tokenizer import (
    XLMRobertaTokenizer as XLMRobertaTokenizer,
)
from keras_hub.src.models.xlnet.xlnet_backbone import (
    XLNetBackbone as XLNetBackbone,
)
from keras_hub.src.tokenizers.tokenizer import Tokenizer as Tokenizer



import numpy as np

from keras_hub.src.models.mixtral.mixtral_backbone import MixtralBackbone
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = MixtralBackbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_query_heads": transformers_config["num_attention_heads"],
        "hidden_dim": transformers_config["hidden_size"],
        "intermediate_dim": transformers_config["intermediate_size"],
        "num_key_value_heads": transformers_config["num_key_value_heads"],
        "num_experts": transformers_config["num_local_experts"],
        "top_k": transformers_config["num_experts_per_tok"],
        "rope_max_wavelength": transformers_config["rope_theta"],
        "layer_norm_epsilon": transformers_config["rms_norm_eps"],
        "sliding_window": transformers_config["sliding_window"],
        "output_router_logits": transformers_config["output_router_logits"],
    }


def convert_weights(backbone, loader, transformers_config):
    # Embeddings
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").embeddings,
        hf_weight_key="model.embed_tokens.weight",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").reverse_embeddings,
        hf_weight_key="lm_head.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
    )

    def transpose_and_reshape(x, shape):
        return np.reshape(np.transpose(x), shape)

    for i in range(backbone.num_layers):
        decoder_layer = backbone.get_layer(f"transformer_layer_{i}")

        # Input layernorm
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layernorm.scale,
            hf_weight_key=f"model.layers.{i}.input_layernorm.weight",
        )

        # Attention layers
        ## Query
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.query_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.q_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Key
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.key_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.k_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Value
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.value_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.v_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Output
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.output_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.o_proj.weight",
            hook_fn=transpose_and_reshape,
        )

        # MoE layers
        # Router gate
        loader.port_weight(
            keras_variable=decoder_layer._sparse_moe_block._sparse_feedforward_gate_dense.kernel,
            hf_weight_key=f"model.layers.{i}.block_sparse_moe.gate.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        # Batched experts: w1 (gate), w3 (intermediate), and w2 (output) weights
        gate_weights_list = []
        intermediate_weights_list = []
        output_weights_list = []
        for expert_idx in range(backbone.num_experts):
            # Load w1 (gate dense) for each expert
            w1 = loader.get_tensor(
                f"model.layers.{i}.block_sparse_moe.experts.{expert_idx}.w1.weight"
            )
            w1_transposed = np.transpose(w1, axes=(1, 0))
            gate_weights_list.append(w1_transposed)

            w3 = loader.get_tensor(
                f"model.layers.{i}.block_sparse_moe.experts.{expert_idx}.w3.weight"
            )
            w3_transposed = np.transpose(w3, axes=(1, 0))
            intermediate_weights_list.append(w3_transposed)

            w2 = loader.get_tensor(
                f"model.layers.{i}.block_sparse_moe.experts.{expert_idx}.w2.weight"
            )
            w2_transposed = np.transpose(w2, axes=(1, 0))
            output_weights_list.append(w2_transposed)

        gate_batched = np.stack(gate_weights_list, axis=0)
        intermediate_batched = np.stack(intermediate_weights_list, axis=0)
        output_batched = np.stack(output_weights_list, axis=0)

        # Assign batched weights to expert_bank
        decoder_layer._sparse_moe_block.expert_bank._expert_feedforward_gate_dense.assign(
            gate_batched
        )
        decoder_layer._sparse_moe_block.expert_bank._expert_feedforward_intermediate_dense.assign(
            intermediate_batched
        )
        decoder_layer._sparse_moe_block.expert_bank._expert_feedforward_output_dense.assign(
            output_batched
        )

        # Feedforward layernorm
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_layernorm.scale,
            hf_weight_key=f"model.layers.{i}.post_attention_layernorm.weight",
        )

    # Final normalization layer
    loader.port_weight(
        keras_variable=backbone.get_layer("sequence_output_layernorm").scale,
        hf_weight_key="model.norm.weight",
    )

    return backbone


def convert_tokenizer(cls, preset, **kwargs):
    return cls(get_file(preset, "tokenizer.model"), **kwargs)



import keras

from keras_hub.src.models.mobilenet.util import adjust_channels


class DropPath(keras.layers.Layer):
    """Implements the DropPath layer.

    DropPath is a form of stochastic depth, where connections are randomly
    dropped during training.

    Args:
        drop_prob: float. The probability of dropping a path.
        scale_by_keep: bool. If `True`, scale the output by `1/keep_prob`.
    """

    def __init__(self, drop_prob=0.0, scale_by_keep=True, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def call(self, x, training=False):
        if self.drop_prob == 0.0 or not training:
            return x
        keep_prob = 1.0 - self.drop_prob
        shape = (keras.ops.shape(x)[0],) + (1,) * (len(x.shape) - 1)
        random_tensor = keep_prob + keras.random.uniform(
            shape, 0, 1, dtype=x.dtype
        )
        random_tensor = keras.ops.floor(random_tensor)
        if keep_prob > 0.0 and self.scale_by_keep:
            random_tensor = random_tensor / keep_prob
        return x * random_tensor

    def get_config(self):
        config = super().get_config()
        config.update(
            {"drop_prob": self.drop_prob, "scale_by_keep": self.scale_by_keep}
        )
        return config


class LayerScale2d(keras.layers.Layer):
    """A layer that applies a learnable scaling factor to the input tensor.

    This layer scales the input tensor by a learnable `gamma` parameter. The
    scaling is applied channel-wise.

    Args:
        dim: int. The number of channels in the input tensor.
        init_values: float. The initial value for the `gamma` parameter.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        dim,
        init_values=1e-5,
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.dim = dim
        self.init_values = init_values
        self.data_format = data_format
        self.channel_axis = channel_axis

    def build(self, input_shape):
        self.gamma = self.add_weight(
            shape=(self.dim,),
            initializer=keras.initializers.Constant(self.init_values),
            trainable=True,
            name="gamma",
        )
        super().build(input_shape)

    def call(self, x):
        if self.data_format == "channels_first":
            gamma = keras.ops.reshape(self.gamma, (1, self.dim, 1, 1))
        else:
            gamma = keras.ops.reshape(self.gamma, (1, 1, 1, self.dim))
        return x * gamma

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "dim": self.dim,
                "init_values": self.init_values,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config


class RmsNorm2d(keras.layers.Layer):
    """A layer that applies Root Mean Square Normalization to a 2D input.

    This layer normalizes the input tensor along the channel dimension using
    the root mean square of the values, and then scales it by a learnable
    `gamma` parameter.

    Args:
        dim: int. The number of channels in the input tensor.
        eps: float. A small epsilon value to avoid division by zero.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        dim,
        eps=1e-6,
        data_format=None,
        channel_axis=None,
        gamma_initializer="ones",
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.dim = dim
        self.eps = eps
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.gamma_initializer = gamma_initializer

    def build(self, input_shape):
        self.gamma = self.add_weight(
            shape=(self.dim,),
            initializer=self.gamma_initializer,
            trainable=True,
            name="gamma",
        )
        super().build(input_shape)

    def call(self, x):
        input_dtype = x.dtype
        if self.data_format == "channels_first":
            x_permuted = keras.ops.transpose(x, (0, 2, 3, 1))
        else:
            x_permuted = x
        x_float = keras.ops.cast(x_permuted, "float32")
        norm_factor = keras.ops.rsqrt(
            keras.ops.mean(keras.ops.square(x_float), axis=-1, keepdims=True)
            + self.eps
        )
        norm_x_float = x_float * norm_factor
        norm_x = keras.ops.cast(norm_x_float, input_dtype)
        scaled_x = norm_x * self.gamma
        if self.data_format == "channels_first":
            output = keras.ops.transpose(scaled_x, (0, 3, 1, 2))
        else:
            output = scaled_x
        return output

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "dim": self.dim,
                "eps": self.eps,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class ConvNormAct(keras.layers.Layer):
    """A layer that combines convolution, normalization, and activation.

    This layer provides a convenient way to create a sequence of a 2D
    convolution, a normalization layer, and an activation function.

    Args:
        out_chs: int. The number of output channels.
        kernel_size: int or tuple. The size of the convolution kernel.
        stride: int or tuple. The stride of the convolution.
        dilation: int or tuple. The dilation rate of the convolution.
        groups: int. The number of groups for a grouped convolution.
        bias: bool. If `True`, a bias term is used in the convolution.
        pad_type: str. The type of padding to use. `"same"` or `""` for same
            padding, otherwise valid padding.
        apply_act: bool. If `True`, an activation function is applied.
        act_layer: str. The name of the activation function to use.
        norm_layer: str. The name of the normalization layer to use.
            Supported values are `"batch_norm"` and `"rms_norm"`.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
    """

    def __init__(
        self,
        out_chs,
        kernel_size,
        stride=1,
        dilation=1,
        groups=1,
        bias=False,
        pad_type="same",
        apply_act=True,
        act_layer="relu",
        norm_layer="batch_norm",
        data_format=None,
        channel_axis=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.out_chs = out_chs
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.groups = groups
        self.bias = bias
        self.pad_type = pad_type
        self.apply_act = apply_act
        self.act_layer = act_layer
        self.norm_layer = norm_layer
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.bias_initializer = "zeros"
        padding_mode = "valid"
        if pad_type.lower() == "" or pad_type.lower() == "same":
            padding_mode = "same"

        self.conv = keras.layers.Conv2D(
            out_chs,
            kernel_size,
            strides=stride,
            padding=padding_mode,
            dilation_rate=dilation,
            groups=groups,
            use_bias=bias,
            data_format=self.data_format,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            dtype=self.dtype_policy,
        )

        if norm_layer == "batch_norm":
            self.norm = keras.layers.BatchNormalization(
                axis=self.channel_axis,
                epsilon=1e-5,
                gamma_initializer="ones",
                beta_initializer="zeros",
                dtype=self.dtype_policy,
            )
        elif norm_layer == "rms_norm":
            self.norm = RmsNorm2d(
                out_chs,
                data_format=self.data_format,
                channel_axis=self.channel_axis,
                gamma_initializer="ones",
                dtype=self.dtype_policy,
            )
        else:
            ln_axis = [1, 2, 3]
            if self.data_format == "channels_first":
                ln_axis = [2, 3, 1]
            self.norm = keras.layers.LayerNormalization(
                axis=ln_axis,
                dtype=self.dtype_policy,
            )

        if self.apply_act:
            if act_layer == "gelu":
                self.act = keras.layers.Activation(
                    lambda x: keras.activations.gelu(x, approximate=False),
                    dtype=self.dtype_policy,
                )
            else:
                self.act = keras.layers.Activation(
                    act_layer,
                    dtype=self.dtype_policy,
                )

    def build(self, input_shape):
        self.conv.build(input_shape)
        conv_output_shape = self.conv.compute_output_shape(input_shape)
        self.norm.build(conv_output_shape)
        if self.apply_act:
            self.act.build(conv_output_shape)
        self.built = True

    def call(self, x, training=False):
        x = self.conv(x)
        x = self.norm(x, training=training)
        if self.apply_act:
            x = self.act(x)
        return x

    def compute_output_shape(self, input_shape):
        return self.conv.compute_output_shape(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "out_chs": self.out_chs,
                "kernel_size": self.kernel_size,
                "stride": self.stride,
                "dilation": self.dilation,
                "groups": self.groups,
                "bias": self.bias,
                "pad_type": self.pad_type,
                "apply_act": self.apply_act,
                "act_layer": self.act_layer,
                "norm_layer": self.norm_layer,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config


class SEModule(keras.layers.Layer):
    """Implements the Squeeze-and-Excitation (SE) module.

    The SE module adaptively recalibrates channel-wise feature responses by
    explicitly modeling interdependencies between channels.

    Args:
        channels: int. The number of input channels.
        rd_ratio: float. The reduction ratio for the bottleneck channels.
        rd_channels: int. The number of bottleneck channels. If specified,
            `rd_ratio` is ignored.
        rd_divisor: int. The divisor for rounding the number of bottleneck
            channels.
        add_maxpool: bool. If `True`, max pooling is used in addition to
            average pooling for the squeeze operation.
        bias: bool. If `True`, bias terms are used in the fully connected
            layers.
        act_layer: str. The activation function for the bottleneck layer.
        norm_layer: str. The normalization layer to use.
        data_format: str. The format of the input data, either
            `"channels_last"` or `"channels_first"`.
        channel_axis: int. The axis representing the channels in the input
            tensor.
        gate_layer: str. The gating activation function.
    """

    def __init__(
        self,
        channels,
        rd_ratio=1.0 / 16,
        rd_channels=None,
        rd_divisor=8,
        add_maxpool=False,
        bias=True,
        act_layer="relu",
        norm_layer=None,
        data_format=None,
        channel_axis=None,
        gate_layer="sigmoid",
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.channels = channels
        self.add_maxpool = add_maxpool
        if not rd_channels:
            rd_channels = adjust_channels(
                channels * rd_ratio, rd_divisor, round_limit=0.0
            )
        self.rd_ratio = rd_ratio
        self.rd_channels = rd_channels
        self.rd_divisor = rd_divisor
        self.bias = bias
        self.act_layer_arg = act_layer
        self.kernel_initializer = keras.initializers.VarianceScaling(
            scale=2.0, mode="fan_out", distribution="untruncated_normal"
        )
        self.bias_initializer = "zeros"
        self.norm_layer_arg = norm_layer
        self.gate_layer_arg = gate_layer
        self.data_format = data_format
        self.channel_axis = channel_axis
        self.mean_axis = [2, 3] if data_format == "channels_first" else [1, 2]
        self.fc1 = keras.layers.Conv2D(
            rd_channels,
            kernel_size=1,
            use_bias=bias,
            name="fc1",
            data_format=self.data_format,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            dtype=self.dtype_policy,
        )
        self.bn = (
            keras.layers.BatchNormalization(
                axis=channel_axis, dtype=self.dtype_policy
            )
            if norm_layer
            else (lambda x, training: x)
        )
        self.act = keras.layers.Activation(act_layer, dtype=self.dtype_policy)
        self.fc2 = keras.layers.Conv2D(
            channels,
            kernel_size=1,
            use_bias=bias,
            name="fc2",
            data_format=self.data_format,
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            dtype=self.dtype_policy,
        )
        self.gate = keras.layers.Activation(gate_layer, dtype=self.dtype_policy)

    def build(self, input_shape):
        self.fc1.build(input_shape)
        fc1_output_shape = self.fc1.compute_output_shape(input_shape)
        if hasattr(self.bn, "build"):
            self.bn.build(fc1_output_shape)
        self.act.build(fc1_output_shape)
        self.fc2.build(fc1_output_shape)
        self.built = True

    def call(self, x, training=False):
        x_se = keras.ops.mean(x, axis=self.mean_axis, keepdims=True)
        if self.add_maxpool:
            x_se = 0.5 * x_se + 0.5 * keras.ops.max(
                x, axis=self.mean_axis, keepdims=True
            )
        x_se = self.fc1(x_se)
        x_se = self.bn(x_se, training=training)
        x_se = self.act(x_se)
        x_se = self.fc2(x_se)
        return x * self.gate(x_se)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "channels": self.channels,
                "rd_ratio": self.rd_ratio,
                "rd_channels": self.rd_channels,
                "rd_divisor": self.rd_divisor,
                "add_maxpool": self.add_maxpool,
                "bias": self.bias,
                "act_layer": self.act_layer_arg,
                "norm_layer": self.norm_layer_arg,
                "gate_layer": self.gate_layer_arg,
                "data_format": self.data_format,
                "channel_axis": self.channel_axis,
            }
        )
        return config



import keras
from keras import ops


class Gemma3InterleaveEmbeddings(keras.layers.Layer):
    """Places image embeddings in the correct position in an embedding sequence.

    For Gemma3, images can be in any position in the input sequence. In order
    to do accomplish this, we have image placeholder tokens in the input
    sequence. We fill up these positions with the image embeddings as returned
    by the vision encoder.

    Args:
        num_vision_tokens_per_image: int. Number of soft tokens per image.
    """

    def __init__(self, num_vision_tokens_per_image, dtype=None, **kwargs):
        super().__init__(dtype=dtype, **kwargs)

        self.num_vision_tokens_per_image = num_vision_tokens_per_image

    def call(self, image_embeddings, text_embeddings, vision_indices):
        """
        Integrates image embeddings into a text embedding sequence.

        Args:
            image_embeddings: tensor. Image embeddings as returned by the
                vision encoder (`Gemma3VisionEncoder`, usually). Shape:
                `(batch_size * num_images_per_prompt, `
                `num_vision_tokens_per_image, embedding_dim)`.
            text_embeddings: tensor. Embeddings returned by the text embedding
                layer. Shape: `(batch_size, seq_length, embedding_dim)`.
            vision_indices:  tensor. Indexes into `text_embeddings`, used to
                identify which places are supposed to be replaced by
                `image_embeddings`. Shape:
                `(batch_size,`
                `num_images_per_prompt * num_vision_tokens_per_image)`.

        Returns:
            Tensor of shape `(batch_size, seq_length, embedding_dim)`
            representing the reconstructed embeddings.
        """

        batch_size, seq_length, embedding_dim = ops.shape(text_embeddings)
        # `num_images` will be 0 for text only inputs, and
        # `batch_size * max_images_per_prompt` if images are passed.
        num_images = ops.shape(image_embeddings)[0]

        # Flatten text embeddings, image embeddings and indices.
        flat_text_embeddings = ops.reshape(
            text_embeddings, (batch_size * seq_length, embedding_dim)
        )
        # `flat_image_embeddings` is the `updates` tensor and should be of shape
        # `(num_updates, embedding_dim)`.
        flat_image_embeddings = ops.reshape(
            image_embeddings,
            (
                num_images * self.num_vision_tokens_per_image,
                embedding_dim,
            ),
        )

        # For vision indices, we need to add values such that the indices
        # index into a flattened `text_embeddings`.
        to_add = ops.multiply(
            keras.ops.arange(batch_size, dtype="int32"), seq_length
        )
        to_add = ops.cast(ops.expand_dims(to_add, axis=-1), "int32")
        vision_indices = ops.add(vision_indices, to_add)

        # indices should be of shape `(num_updates, 1)`. `num_updates` is
        # how many vision tokens there are to update.
        vision_indices_shape = ops.shape(vision_indices)
        flat_vision_indices = ops.reshape(
            vision_indices,
            (vision_indices_shape[0] * vision_indices_shape[1], 1),
        )
        indices = ops.cast(flat_vision_indices, "int32")

        # Before reconstructing, store the 0th index so that we can restore it
        # later.
        zeroth_index_text_embeddings = ops.take(
            flat_text_embeddings,
            indices=ops.squeeze(to_add, axis=-1),
            axis=0,
        )

        # Reconstruct embeddings
        reconstructed_embedding = ops.scatter_update(
            inputs=flat_text_embeddings,
            indices=indices,
            updates=flat_image_embeddings,
        )

        # Remember that we pad `vision_indices` with the 0th index. We need to
        # restore the original value in the reconstructed embedding tensor.
        reconstructed_embedding = ops.scatter_update(
            inputs=reconstructed_embedding,
            indices=to_add,
            updates=zeroth_index_text_embeddings,
        )

        # Reshape to original dimensions
        reconstructed_embedding = ops.reshape(
            reconstructed_embedding, (batch_size, seq_length, embedding_dim)
        )
        return reconstructed_embedding

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_vision_tokens_per_image": self.num_vision_tokens_per_image,
            }
        )
        return config



"""DINOV2 model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "dinov2_small": {
        "metadata": {
            "description": (
                "Vision Transformer (small-sized model) trained using DINOv2."
            ),
            "params": 22_582_656,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_small/1",
    },
    "dinov2_base": {
        "metadata": {
            "description": (
                "Vision Transformer (base-sized model) trained using DINOv2."
            ),
            "params": 87_632_640,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_base/1",
    },
    "dinov2_large": {
        "metadata": {
            "description": (
                "Vision Transformer (large-sized model) trained using DINOv2."
            ),
            "params": 305_771_520,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_large/1",
    },
    "dinov2_giant": {
        "metadata": {
            "description": (
                "Vision Transformer (giant-sized model) trained using DINOv2."
            ),
            "params": 1_138_585_088,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_giant/1",
    },
    "dinov2_with_registers_small": {
        "metadata": {
            "description": (
                "Vision Transformer (small-sized model) trained using DINOv2, "
                "with registers."
            ),
            "params": 22_584_192,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_with_registers_small/1",
    },
    "dinov2_with_registers_base": {
        "metadata": {
            "description": (
                "Vision Transformer (base-sized model) trained using DINOv2, "
                "with registers."
            ),
            "params": 87_635_712,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_with_registers_base/1",
    },
    "dinov2_with_registers_large": {
        "metadata": {
            "description": (
                "Vision Transformer (large-sized model) trained using DINOv2, "
                "with registers."
            ),
            "params": 305_775_616,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_with_registers_large/1",
    },
    "dinov2_with_registers_giant": {
        "metadata": {
            "description": (
                "Vision Transformer (giant-sized model) trained using DINOv2, "
                "with registers."
            ),
            "params": 1_138_591_232,
            "path": "dinov2",
        },
        "kaggle_handle": "kaggle://keras/dinov2/keras/dinov2_with_registers_giant/1",
    },
}



from keras_hub.src.models.deeplab_v3.deeplab_v3_backbone import (
    DeepLabV3Backbone,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, DeepLabV3Backbone)
